// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cmd

import (
	"encoding/json"
	"errors"
	"fmt"
	"github.com/spf13/cobra"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"io"
	"os"
	"strings"
)

var createCmd = &cobra.Command{
	Use:   "create",
	Short: "Create definitions",
	Long:  "Command Get creates definitions of objects from json files",
	Args: cobra.MinimumNArgs(1),
	RunE: func(cmd *cobra.Command, args []string) error {
		// fetch context with connections to repository
		c, _ := cmd.Context().(*DesignerCtx)
		var err error
		// file representation in struct
		type input struct {
			Modules []model.ModuleDefinition
			Operations []model.OperationDefinition
			Edges []model.EdgeDefinition
			Nodes []model.NodeDefinition
		}
		modules := make([]*model.ModuleDefinition, 0)
		operations := make([]*model.OperationDefinition, 0)
		edges := make([]*model.EdgeDefinition,0)
		nodes := make([]*model.NodeDefinition,0)
		for _, f := range args {
			prefixErrorFile := fmt.Sprintf("error read from file %s", f)
			var (
				input input
			)
			file, err := os.Open(f)
			if err != nil {
				return fmt.Errorf("%s: %s", prefixErrorFile, err)
			}
			dec := json.NewDecoder(file)
			dec.DisallowUnknownFields()
			err = dec.Decode(&input)
			if err != nil {
				var syntaxError *json.SyntaxError
				var unmarshalTypeError *json.UnmarshalTypeError
				switch {
				case errors.As(err, &syntaxError):
					return fmt.Errorf("%s: badly-formed JSON (at position %d)",prefixErrorFile, syntaxError.Offset)
				case errors.Is(err, io.ErrUnexpectedEOF):
					return fmt.Errorf("%s: badly-formed JSON", prefixErrorFile)
				case errors.As(err, &unmarshalTypeError):
					return fmt.Errorf("%s: invalid value for the %q field (at position %d)", prefixErrorFile, unmarshalTypeError.Field, unmarshalTypeError.Offset)
				case strings.HasPrefix(err.Error(), "json: unknown field "):
					fieldName := strings.TrimPrefix(err.Error(), "json: unknown field ")
					return fmt.Errorf("%s: unknown field '%s'", prefixErrorFile, fieldName)
				}
			}
			for _, m := range input.Modules {
				mod := m
				modules = append(modules, &mod)
			}
			for _, n := range input.Nodes {
				nod := n
				nodes = append(nodes, &nod)
			}
			for _, e := range input.Edges {
				ed := e
				edges = append(edges, &ed)
			}
			for _, o := range input.Operations {
				op := o
				operations = append(operations, &op)
			}
		}
		// store all to repository
		// todo firstly validate so rollback everything is possible
		err = c.Storer.StoreModule(modules...)
		if err != nil {
			return fmt.Errorf("error storing modules: %s", err)
		}
		err = c.Storer.StoreOperation(operations...)
		if err != nil {
			return fmt.Errorf("error storing operations: %s", err)
		}
		err = c.Storer.StoreNode(nodes...)
		if err != nil {
			return fmt.Errorf("error storing nodes: %s", err)
		}
		err = c.Storer.StoreEdge(edges...)
		if err != nil {
			return fmt.Errorf("error storing edges: %s", err)
		}
		return nil
	},
}


