// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cmd

import (
	"encoding/json"
	"fmt"
	"github.com/spf13/cobra"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"strconv"
)

func init()  {
	getNodeCmd.Flags().StringP( "operation", "","","filter query to nodes of specific operation")
	getNodeCmd.Flags().StringP( "module", "","","filter query to nodes of that contain specific module")
	getEdgeCmd.Flags().StringP( "operation", "","","filter query to edges of specific operation")
	getEdgeCmd.Flags().StringP( "from", "","","filter query to edges starting at specific node")
	getEdgeCmd.Flags().StringP( "to", "","","filter query to edges ending in specific node")
	getModuleCmd.Flags().StringP("type", "", "", "filter query to modules of specific type")
	getModuleCmd.Flags().StringP("node", "", "", "filter query to module that is contained in specific node")
	getCmd.PersistentFlags().IntP("limit", "", 0, "limit number of objects printed (default is unlimited)")
	getCmd.AddCommand(getOperationCmd)
	getCmd.AddCommand(getNodeCmd)
	getCmd.AddCommand(getEdgeCmd)
	getCmd.AddCommand(getModuleCmd)
}

var getCmd = &cobra.Command{
	Use:               "get",
	Short:             "Print definitions",
	Long: 			   "Command Get prints definition of chosen objects.",
}

var getOperationCmd = &cobra.Command{
	Use:   "operation",
	Short: "Print definition of operation",
	RunE: func(cmd *cobra.Command, args []string) error {
		// fetch context with connections to repository
		c, _ := cmd.Context().(*DesignerCtx)
		var (
			err error
			limit int
		)
		// get limit of results
		limit, _ = cmd.Flags().GetInt("limit")
		ops := make([]*model.OperationDefinition, 0)
		if len(args) == 0 {
			// no operation specified, fetch them all
			ops, err = c.Fetcher.QueryOperation(nil, limit)
			if err != nil {
				return fmt.Errorf("error fetching from repository: %s", err)
			}
		} else {
			// fetch only specified operations
			for _, name := range args {
				o, err := c.Fetcher.QueryOperation(&model.OperationDefinition{Name: name}, limit)
				if err != nil {
					return fmt.Errorf("error fetching from repository: %s", err)
				}
				ops = append(ops, o...)
			}
		}
		// print all fetched operations
		for _, o := range ops {
			out, err := json.MarshalIndent(o, "", "   ")
			if err != nil {
				fmt.Println("internal error during data serialize")
			} else {
				fmt.Println(string(out))
			}
		}
		return nil
	},
}

var getNodeCmd = &cobra.Command{
	Use:   "node",
	Short: "Print definition of node",
	RunE: func(cmd *cobra.Command, args []string) error {
		// fetch context with connections to repository
		c, _ := cmd.Context().(*DesignerCtx)
		var (
			err error
			operationName string
			moduleName string
			limit int
		)
		// get limit of results
		limit, _ = cmd.Flags().GetInt("limit")
		// get module name from flag if used
		moduleName, _ = cmd.Flags().GetString("module")
		// get operation name from flag if used
		operationName, _ = cmd.Flags().GetString("operation")
		nodes := make([]*model.NodeDefinition, 0)
		// if no node specified, fetch them all
		if len(args) == 0 {
			nodes, err = c.Fetcher.QueryNode(&model.NodeDefinition{OperationName: operationName, ModuleName: moduleName}, limit)
			if err != nil {
				return fmt.Errorf("error fetching from repository: %s", err)
			}
		// fetch only specified node
		} else {
			for _, name := range args {
				var id int
				var err error
				if id, err = strconv.Atoi(name); err != nil || uint(id) == 0 {
					// node with definition id other than number doesnt exist
					continue
				}
				node, err := c.Fetcher.QueryNode(&model.NodeDefinition{ID: uint(id), OperationName: operationName, ModuleName: moduleName}, limit)
				if err != nil {
					return fmt.Errorf("error fetching from repository: %s", err)
				}
				nodes = append(nodes, node...)
			}
		}
		// print all fetched operations
		for _, n := range nodes {
			out, err := json.MarshalIndent(n, "", "   ")
			if err != nil {
				fmt.Println("internal error during data serialize")
			} else {
				fmt.Println(string(out))
			}
		}
		return nil
	},
}

var getEdgeCmd = &cobra.Command{
	Use:   "edge",
	Short: "Print definition of operation",
	Args: cobra.NoArgs,
	RunE: func(cmd *cobra.Command, args []string) error {
		// fetch context with connections to repository
		c, _ := cmd.Context().(*DesignerCtx)
		var (
			err error
			limit int
			from int
			to int
			operation string
		)
		// get limit of results
		limit, _ = cmd.Flags().GetInt("limit")
		// get limit of results
		f, _ := cmd.Flags().GetString("from")
		if from, err = strconv.Atoi(f); err != nil {
			return nil
		}
		// get limit of results
		t, _ := cmd.Flags().GetString("to")
		if to, err = strconv.Atoi(t); err != nil {
			return nil
		}
		// get limit of results
		operation, _ = cmd.Flags().GetString("operation")

		edges := make([]*model.EdgeDefinition, 0)
		edges, err = c.Fetcher.QueryEdge(&model.EdgeDefinition{OperationName: operation, FromNode: uint(from), ToNode: uint(to)}, limit)
		if err != nil {
				return fmt.Errorf("error fetching from repository: %s", err)
			}
		// print all fetched operations
		for _, e := range edges {
			out, err := json.MarshalIndent(e, "", "   ")
			if err != nil {
				fmt.Println("internal error during data serialize")
			} else {
				fmt.Println(string(out))
			}
		}
		return nil
	},
}

var getModuleCmd = &cobra.Command{
	Use:   "module",
	Short: "Print definition of module",
	RunE: func(cmd *cobra.Command, args []string) error {
		c, _ := cmd.Context().(*DesignerCtx)
		var (
			err error
			limit int
			moduleType string
			node int
			moduleName string
		)
		// get limit of results
		limit, _ = cmd.Flags().GetInt("limit")
		// get limit of results
		moduleType, _ = cmd.Flags().GetString("type")
		// get limit of results
		n, _ := cmd.Flags().GetString("node")
		if n != "" {
			if node, err = strconv.Atoi(n); n != "" && err != nil {
				return nil
			}
			nodes, err := c.Fetcher.QueryNode(&model.NodeDefinition{ID: uint(node)}, 1)
			if err != nil {
				return fmt.Errorf("error fetching from repository: %s", err)
			}
			if len(nodes) == 1 {
				moduleName = nodes[0].ModuleName
			}
		}
		modules := make([]*model.ModuleDefinition, 0)
		// no operation specified, fetch them all
		if len(args) == 0 || moduleName != "" {
			modules, err = c.Fetcher.QueryModule(&model.ModuleDefinition{Name: moduleName, Type: model.ModuleType(moduleType)}, limit)
			if err != nil {
				return fmt.Errorf("error fetching from repository: %s", err)
			}
			// fetch only specified operations
		} else {
			for _, name := range args {
				m, err := c.Fetcher.QueryModule(&model.ModuleDefinition{Name: name, Type: model.ModuleType(moduleType)}, limit)
				if err != nil {
					return fmt.Errorf("error fetching from repository: %s", err)
				}
				modules = append(modules, m...)
			}
		}
		// print all fetched operations
		for _, m := range modules {
			out, err := json.MarshalIndent(m, "", "   ")
			if err != nil {
				fmt.Println("internal error during data serialize")
			} else {
				fmt.Println(string(out))
			}
		}
		return nil
	},
}