// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package app

import (
	"errors"
	"fmt"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/pkg/mEngine"
	"strconv"
	"strings"
	"sync"
	"time"
)

type ErrorBadRequest struct {
	Msg    string
}

func (e *ErrorBadRequest) Error() string {
	return e.Msg
}

type ErrorInternal struct {
	Msg    string
}

func (e *ErrorInternal) Error() string {
	return e.Msg
}

// Event represents something that should be done next. Event Service process all events.
type Event interface {
	// private method so only orchestrator can define events
	private()
	// register method will provide orchestrator service methods to event
	register(h *orchestratorService)
	// validates event
	validate() error
	// prepares event for handling
	prepare() error
	// start event processing
	handle() error
	// start critical part of event processing
	critical()
}

// Event to delete already running operation
type DeleteOperationEvent struct {
	OperationID string

	// internal metadata used at event process
	operation *model.Operation

	handler *orchestratorService
}

// Event to create new operation workflow and run it.
type StartOperationEvent struct {
	// operation id that user chosen
	ID string
	// name of definition to use
	Definition string
	// input parameters
	Parameters map[string]map[string]string

	// internal metadata used at event process
	operation  *model.Operation
	definition *model.OperationDefinition

	response string

	handler *orchestratorService
}

func (e *StartOperationEvent) validate() error {
	if e.Definition == "" {
		return &ErrorBadRequest{Msg: "definition not provided"}
	}
	if len(e.ID) != 0 {
		if !strings.ContainsAny(e.ID[:1], "qwertyuiopasdfghjklzxcvbnm") {
			return &ErrorBadRequest{Msg: "provided ID " + e.ID + "does not meet requirements: must start with small letter"}
		}
	}
	return nil
}

func (e *StartOperationEvent) prepare() error {
	handler := e.handler
	if e.ID == "" {
		e.ID = handler.genUUID("operation-")
	}
	// find definition
	def, err := handler.definition.GetOperation(&model.OperationDefinition{Name: e.Definition}, 1)
	if err != nil  {
		return &ErrorBadRequest{Msg: err.Error()}
	} else if len(def) == 0 {
		return &ErrorBadRequest{Msg: "definition with name " + e.Definition + "not found"}
	}
	// transfer input parameters
	parameters := make([]model.OperationParameter, 0)
	for module, moduleParams := range e.Parameters {
		for key, value := range moduleParams {
			parameters = append(parameters, model.OperationParameter{
				ModuleName:  module,
				Key:         key,
				Value:       value,
			})
		}
	}
	// create nodes based on definition
	nodeIDs := make(map[uint]string)
	nodes := make([]model.Node, 0)
	for _, node := range def[0].Nodes {
		nodeIDs[node.ID] = handler.genUUID("node-")
		nodes = append(nodes, model.Node{
			ID:	nodeIDs[node.ID],
			Definition:  node.ID,
		})
	}
	// create edges based on definition
	edges := make([]model.Edge, 0)
	for _, edge := range def[0].Edges {
		edges = append(edges, model.Edge{
			FromNode:    nodeIDs[edge.FromNode],
			ToNode:      nodeIDs[edge.ToNode],
		})
	}

	op, err := handler.metadata.AddOperation(&model.Operation{
		ID:              e.ID,
		Definition:      e.Definition,
		InputParameters: parameters,
		Status:          model.OperationStatusScheduling,
		StatusReason:    "",
		Nodes:           nodes,
		Edges:           edges,
	})
	if err != nil {
		return &ErrorBadRequest{Msg: fmt.Sprintf("cannot create operation: %s", err)}
	}
	e.operation = op
	e.definition = def[0]
	e.response = e.ID
	return nil
}

func (e *StartOperationEvent) handle() error {
	err := e.handler.metadata.PrepareInputParameters(e.operation)
	if err != nil {
		_, _, _ = e.handler.metadata.SetOperationStatus(e.operation.ID, model.OperationStatusFailed, err.Error())
		return err
	}
	return nil
}

func (e *StartOperationEvent) critical() {
	handler := e.handler

	// define which module will be used at which node
	handler.assignModules(e.operation, e.definition)
	// transform parameters into searchable structure
	params := make(map[string]map[string]string)
	for _, p := range e.operation.InputParameters {
		if params[p.ModuleName] == nil {
			params[p.ModuleName] = make(map[string]string)
		}
		params[p.ModuleName][p.Key] = p.Value
	}
	// create tasks to schedule run every node
	for _, node := range e.operation.Nodes {
		// schedule only nodes that are not already running
		nodeID := node.ID
		moduleID := *node.ModuleID
		module, _ := handler.metadata.GetModule(&model.Module{ID: moduleID}, 1)
		if module[0].Status != model.ModuleStatusScheduling {
			continue
		} else {
			if status, _ := handler.metadata.SetModuleStatus(moduleID, model.ModuleStatusCreating);
				status != model.ModuleStatusCreating {
				// todo handle error
				continue
			}
		}
		// find parameters of modules in subtree
		modules := handler.modulesFromSubtree(e.operation, nodeID)
		// make env
		env := make(map[string]string)
		// todo save keywords and do validation at event register
		env["DATAWAREHOUSE"] = handler.config.Warehouse
		env["MODULE_ID_SELF"] = moduleID
		env["MODULE_NAME"] = module[0].Definition
		for _, m := range modules {
			modName := strings.ReplaceAll(strings.ToUpper(m.Definition), "-", "_")
			env["MODULE_ID_" + modName] = m.ID
			for _, p := range m.Parameters {
				keyName := strings.ReplaceAll(strings.ToUpper(p.Key), "-", "_")
				env["PARAMETER_" + modName + "_" + keyName] = p.Value
			}
			for _, p := range m.Ports {
				portName := strings.ReplaceAll(strings.ToUpper(p.Name), "-", "_")
				moduleDNSname := m.ID
				if p.IsPublic {
					moduleDNSname = m.ID + "-external"
				}
				env["ENDPOINT_" + modName + "_" + portName] = moduleDNSname + ":" + strconv.Itoa(p.Number)
			}
		}

		taskOperationCreate := func(doneFunc mEngine.DoneFunc) {
			defer close(doneFunc())
			var err error
			module, _ := handler.metadata.GetModule(&model.Module{ID: moduleID}, 1)
			// todo handle error
			definition, _ := handler.definition.GetModule(&model.ModuleDefinition{Name: module[0].Definition}, 1)
			if handler.config.DockerRegistry != "" {
				definition[0].Image = handler.config.DockerRegistry + "/" + definition[0].Image
			}
			// todo handle error
			if status := handler.getModuleStatus(moduleID); status == model.ModuleStatusFailed {
				return
			}
			module[0], err = handler.platform.StartModule(module[0], definition[0], env)
			if err != nil {
				e.handler.logger.Log("error", err)
				_, _, _ = handler.metadata.SetOperationStatus(e.operation.ID, model.OperationStatusFailed,
					fmt.Sprintf("module %s failed to start", moduleID))
				return
			}
			_ = handler.metadata.ModuleAfterStart(module[0])
			return
		}
		// for each previous node create task that blocks starting new module until node is started
		blockers := make([]mEngine.Blocker, 0)
		for _,n := range handler.previousNodes(e.operation, nodeID) {
			mID := *n.ModuleID
			blockers = append(blockers, func(stopFunc mEngine.StopFunc, doneFunc mEngine.DoneFunc) {
				defer close(doneFunc())
				for {
					select {
					case <-stopFunc():
						return
					case <- time.After(time.Millisecond * 700):
						if status := handler.getModuleStatus(moduleID); status == model.ModuleStatusFailed {
							return
						}
						// set status of node to failed if this predecessor failed
						switch handler.getModuleStatus(mID) {
						case model.ModuleStatusCreating, model.ModuleStatusScheduling:
							continue
						case model.ModuleStatusFailed, model.ModuleStatusUnknown:
							_,_ = handler.metadata.SetModuleStatus(moduleID, model.ModuleStatusFailed)
							_, _, _ = handler.metadata.SetOperationStatus(e.operation.ID, model.OperationStatusFailed,
								fmt.Sprintf("module %s failed", mID))
							return
						case model.ModuleStatusRunning:
							m, _ := handler.metadata.GetModule(&model.Module{ID: mID},1)
							def, _ := handler.definition.GetModule(&model.ModuleDefinition{Name: m[0].Definition}, 1)
							if def[0].IsBlocking {
								continue
							} else {
								return
							}
						default:
							return
						}
					}
				}
			})
		}
		// create blockers
		handler.mEngine.NewTask(moduleID, taskOperationCreate, blockers...)
		_, _, _ = handler.metadata.SetOperationStatus(e.operation.ID, model.OperationStatusRunning, "")
		// todo handling errors
	}
	return
}

func (e *DeleteOperationEvent) validate() error {
	ops, err := e.handler.metadata.GetOperation(&model.Operation{ID: e.OperationID}, 1)
	if err != nil {
		return err
	}
	if len(ops) != 0 {
		e.operation = ops[0]
	} else {
		return errors.New("operation not found")
	}
	return nil
}

func (e *DeleteOperationEvent) prepare() error {
	_, notAlreadyDeleting, err := e.handler.metadata.SetOperationStatus(e.OperationID, model.OperationStatusDeleting, "")
	if err != nil {
		return err
	}
	if !notAlreadyDeleting {
		return errors.New("operation already scheduled to delete")
	}
	return nil
}

func (e *DeleteOperationEvent) handle() error {
	return nil
}

func (e *DeleteOperationEvent) critical() {
	handler := e.handler

	// used in final task to wait for all module delete tasks to finish
	waiter := sync.WaitGroup{}
	// used in final task to determine if all delete tasks succeeded
	deleteWasOk := true
	for _, n := range e.operation.Nodes {
		node := n
		// skip node that has no module assigned
		if node.ModuleID == nil || *node.ModuleID == "" {
			handler.logger.Log("info", fmt.Sprintf("delete operation %s node %s: no module assigned..skipping", e.OperationID, node.ID))
			continue
		}
		// find if module has more nodes assigned
		module, err := handler.metadata.GetModule(&model.Module{ID: *node.ModuleID} ,1)
		if err != nil {
			handler.logger.Log("error", fmt.Sprintf("delete operation %s node %s: %s", e.OperationID, node.ID, err))
			continue
		}
		if len(module) == 0 {
			handler.logger.Log("error", fmt.Sprintf("delete operation %s node %s: assigned module do not exist", e.OperationID, node.ID))
			continue
		}
		if len(module[0].InNodes) > 1 {
			err := handler.metadata.AssignModule("", n.ID)
			if err != nil {
				handler.logger.Log("error", fmt.Sprintf("delete operation %s node %s: %s", e.OperationID, node.ID, err))
			}
			continue
		} else {
			// set module status to deleting
			_, _ = handler.metadata.SetModuleStatus(module[0].ID, model.ModuleStatusDeleting)
			// todo check if error handle needed
		}
		waiter.Add(1)
		// prepare delete task on mEngine
		taskModuleDelete := func(doneFunc mEngine.DoneFunc) {
			defer close(doneFunc())
			defer waiter.Done()
			// stop running module on platform
			err := handler.platform.StopModule(*node.ModuleID)
			if err != nil {
				// log err
				handler.logger.Log("error", fmt.Sprintf("delete operation %s node %s: cannot delete module: %s ", e.OperationID, node.ID, err))
				_, _ = handler.metadata.SetModuleStatus(*node.ModuleID, model.ModuleStatusDeleteFailed)
				deleteWasOk = false
				return
			}
			_ = handler.metadata.repository.DeleteModule(&model.Module{ID: *node.ModuleID})
		}
		// schedule delete task on mEngine
		handler.mEngine.NewTaskWithCancel(*n.ModuleID, taskModuleDelete)
	}
	// delete operation when every module deleted
	taskOperationDelete := func(doneFunc mEngine.DoneFunc) {
		defer close(doneFunc())
		// wait for all deleting tasks to finish
		waiter.Wait()
		// get results of delete modules
		if !deleteWasOk {
			_, _, _ = handler.metadata.SetOperationStatus(e.OperationID, model.OperationStatusDeleteFailed, "")
			return
		}
		err := handler.metadata.repository.DeleteOperation(&model.Operation{ID: e.OperationID})
		if err != nil {
			handler.logger.Log("error", fmt.Sprintf("delete operation %s: %s", e.OperationID, err))
		}
	}
	handler.mEngine.NewTask("operation-" + e.OperationID, taskOperationDelete)
	return
}

func (e *StartOperationEvent) register(h *orchestratorService) {
	e.handler = h
}

func (e *DeleteOperationEvent) register(h *orchestratorService) {
	e.handler = h
}

func (e *StartOperationEvent) Response() string {
	return e.response
}

// to satisfy Event interface
func (StartOperationEvent) private() {}

// to satisfy Event interface
func (DeleteOperationEvent) private() {}

