// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package app

import (
	"fmt"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/logger"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
)

// metadataService provides methods for correct population and handling of metadata structures
// Both repositories must be provided in order for service to work.
type metadataService struct {
	definitions model.DefinitionFetcher
	repository  model.MetadataManager
	logger      logger.Logger
}

// GetOperation returns a number of operations that satisfy provided query.
// Parameter limit sets a maximum number of operations fetched.
// All operations satisfy nil query
func (s *metadataService) GetOperation(query *model.Operation, limit int) ([]*model.Operation, error) {
	return s.repository.QueryOperation(query, limit)
}

// GetNode returns a number of nodes that satisfy provided query.
// Parameter limit sets a maximum number of nodes fetched.
// All nodes satisfy nil query
func (s *metadataService) GetNode(query *model.Node, limit int) ([]*model.Node, error) {
	return s.repository.QueryNode(query, limit)
}

// GetEdge returns a number of edges that satisfy provided query.
// Parameter limit sets a maximum number of edges fetched.
// All edges satisfy nil query
func (s *metadataService) GetEdge(query *model.Edge, limit int) ([]*model.Edge, error) {
	return s.repository.QueryEdge(query, limit)
}

// GetModule returns a number of modules that satisfy provided query.
// Parameter limit sets a maximum number of modules fetched.
// All modules modules nil query
func (s *metadataService) GetModule(query *model.Module, limit int) ([]*model.Module, error) {
	return s.repository.QueryModule(query, limit)
}


// GetNodePredecessor finds predecessor to node with specific definition of module used or nil if not found.
func (s *metadataService) GetNodePredecessor(nodeID string, moduleDefinition string) *model.Node {
	if nodeID == "" || moduleDefinition == "" {
		return nil
	}
	edges, err := s.repository.QueryEdge(&model.Edge{ToNode: nodeID}, 1)
	if err != nil || len(edges) == 0 {
		// log
		return nil
	}
	for _, e := range edges {
		node, err := s.repository.QueryNode(&model.Node{
			ID: e.FromNode,
		}, 1)
		if err != nil || len(node) != 1 {
			continue
		}
		def, err := s.definitions.QueryNode(&model.NodeDefinition{ID: node[0].Definition}, 0)
		if err != nil || len(def) != 1 {
			continue
		}
		if def[0].ModuleName != moduleDefinition {
			continue
		}
		return node[0]
	}
	return nil
}

// GetNodeSuccessor finds successor to node with specific definition of module used or nil if not found.
func (s *metadataService) GetNodeSuccessor(nodeID string, moduleDefinition string) *model.Node {
	if nodeID == "" || moduleDefinition == "" {
		return nil
	}
	edges, err := s.repository.QueryEdge(&model.Edge{FromNode: nodeID}, 1)
	if err != nil || len(edges) == 0 {
		// log
		return nil
	}
	for _, e := range edges {
		node, err := s.repository.QueryNode(&model.Node{
			ID: e.FromNode,
		}, 1)
		if err != nil || len(node) != 1 {
			continue
		}
		def, err := s.definitions.QueryNode(&model.NodeDefinition{ID: node[0].Definition}, 0)
		if err != nil || len(def) != 1 {
			continue
		}
		if def[0].ModuleName != moduleDefinition {
			continue
		}
		return node[0]
	}
	return nil
}

// validate and prepare newly created operation
func (s *metadataService) PrepareOperation(operation *model.Operation) error {
	return s.PrepareInputParameters(operation)
}

// Validate input parameters of operation and append defined parameter defaults
func (s *metadataService) PrepareInputParameters(operation *model.Operation) error {
	// fetch definitions of parameters for particular operation
	nodeParameters :=  make(map[string]map[string]model.NodeDefinitionParameter, 0)
	moduleParameters := make(map[string]map[string]model.ModuleDefinitionParameter, 0)
	inputParameters := make(map[string]map[string]string)
	resultParameters := make([]model.OperationParameter, 0)
	for _, n := range operation.Nodes {
		nodeDef, err := s.definitions.QueryNode(&model.NodeDefinition{ID: n.Definition, OperationName: operation.Definition}, 1)
		if len(nodeDef) != 1 || err != nil {
			return fmt.Errorf("input parameter validation: %s", err)
		}
		moduleDef, err := s.definitions.QueryModule(&model.ModuleDefinition{Name: nodeDef[0].ModuleName}, 1)
		if len(moduleDef) != 1 || err != nil {
			return fmt.Errorf("input parameter validation: %s", err)
		}
		// initialise maps for module
		moduleParameters[moduleDef[0].Name] = make(map[string]model.ModuleDefinitionParameter)
		nodeParameters[moduleDef[0].Name] = make(map[string]model.NodeDefinitionParameter)
		inputParameters[moduleDef[0].Name] = make(map[string]string)
		// make node parameters searchable
		for _, p := range nodeDef[0].Parameters {
			nodeParameters[moduleDef[0].Name][p.Name] = p
		}
		// make module parameters searchable
		for _, p := range moduleDef[0].Parameters {
			moduleParameters[moduleDef[0].Name][p.Name] = p
		}
	}
	// make input parameters searchable
	for _, param := range operation.InputParameters {
		if inputParameters[param.ModuleName] == nil {
			inputParameters[param.ModuleName] = make(map[string]string)
		}
		inputParameters[param.ModuleName][param.Key] = param.Value
	}

	// validate and populate parameters for all module definitions
	for module, moduleParams := range moduleParameters {
		nodeParams := nodeParameters[module]
		inputParams := inputParameters[module]
		// validate all parameters of particular module
		for key, moduleDefaults := range moduleParams {
			result := model.OperationParameter{
				ModuleName: module,
				Key: key,
			}
			_, defined := inputParams[key]
			if !defined {
				// parameter not defined in input so put defaults either from module or node
				if p, ok := nodeParams[key]; ok && p.DefaultValue != nil {
					result.Value = *p.DefaultValue
				} else {
					result.Value = moduleDefaults.DefaultValue
				}
			} else { // parameter defined
				result.Value = inputParams[key]
				// check if user trying to change constant
				if p, ok := nodeParams[key]; ok && p.IsConstant && *p.DefaultValue != result.Value {
					return fmt.Errorf("input parameter validation: %s.%s defined as constant", module, key)
				}
			}
			resultParameters = append(resultParameters, result)
		}
	}
	// assign populated parameters back to operation
	operation.InputParameters = resultParameters
	_ = s.repository.UpdateOperation(&model.Operation{InputParameters: resultParameters})
	return nil
}

// Deletes operation from repository including its nodes and edges
func (s *metadataService) DeleteOperation(operation *model.Operation) error {
	if operation == nil {
		return nil
	}
	err := s.repository.DeleteOperation(operation)
	if err != nil {
		return fmt.Errorf("operation delete: %s", err)
	}
	return nil
}

// Adds operation to repository including its nodes and edges
func (s *metadataService) AddOperation(operation *model.Operation) (*model.Operation, error) {
	if operation == nil {
		return nil, nil
	}
	op, err := s.repository.StoreOperation(operation)
	if err != nil {
		return op, fmt.Errorf("operation add: %s", err)
	}
	return op, nil
}

// assigns module to node, if module not specified, it removes assigned one
func (s *metadataService) AssignModule(moduleID, nodeID string) error {
	if nodeID == "" {
		return fmt.Errorf("module assign: ID of node not provided")
	}
	err := s.repository.UpdateNode(&model.Node{ID: nodeID, ModuleID: &moduleID})
	if err != nil {
		return fmt.Errorf("module assign: %s", err)
	}
	return nil
}

// delete module from repository
func (s *metadataService) DeleteModule(module *model.Module) error {
	//if module.ID == "" {
	//	// no need to delete anything
	//	return nil
	//}
	err := s.repository.DeleteModule(module)
	if err != nil {
		return fmt.Errorf("module delete: %s", err)
	}
	return nil
}

// add module to repository
func (s *metadataService) AddModule(module *model.Module) (*model.Module, error) {
	if module == nil {
		return nil, nil
	}
	m, err := s.repository.StoreModule(module)
	if err != nil {
		return m, fmt.Errorf("module add: %s", err)
	}
	return m, nil
}

// checks if already running module A can be reused as module B
func (s *metadataService) IsModuleReusable(A model.Module, B model.Module) bool {
	// check if modules have same definition
	if A.Definition != B.Definition {
		return false
	}
	running := make(map[string]model.ModuleParameter)
	definitions := make(map[string]model.ModuleDefinitionParameter)
	def, err := s.definitions.QueryModule(&model.ModuleDefinition{Name: A.Definition}, 1)
	if len(def) != 1 || err != nil {
		// todo log
		return false
	}
	// make params of running module and its definition searchable
	for _, p := range def[0].Parameters {
		definitions[p.Name] = p
	}
	for _, p := range A.Parameters {
		running[p.Key] = p
	}
	// check all input params
	for _, p := range B.Parameters {
		// check if param should exist by definition
		if _, ok := definitions[p.Key]; !ok {
			return false
		}
		// if parameter checking is set to strict, check if both are same
		if definitions[p.Key].IsStrict && p.Value != running[p.Key].Value {
			return false
		}
	}
	return true
}

// SetModuleStatus tries to update status of running module if its possible to rules.
// Returns set status of module and occurred error
func (s *metadataService) SetModuleStatus(moduleID string, status model.ModuleStatus) (model.ModuleStatus, error) {
	if moduleID == "" {
		return model.ModuleStatusUnknown, fmt.Errorf("set module status: no ID provided")
	}
	m, err := s.repository.QueryModule(&model.Module{ID: moduleID}, 1)
	if err != nil {
		return model.ModuleStatusUnknown, fmt.Errorf("set module status: %s", err)
	}
	if len(m) == 0 {
		return model.ModuleStatusUnknown, fmt.Errorf("set module status: module %s not found", moduleID)
	}
	if status == m[0].Status {
		return status, nil
	}
	// rules to avoid state inconsistency
	switch m[0].Status {
	case model.ModuleStatusFailed:
		if status != model.ModuleStatusDeleting {
			return model.ModuleStatusFailed, nil
		}
	case model.ModuleStatusDeleting:
		if status != model.ModuleStatusDeleteFailed {
			return model.ModuleStatusDeleting, nil
		}
	case model.ModuleStatusDeleteFailed:
		if status != model.ModuleStatusDeleting {
			return model.ModuleStatusDeleteFailed, nil
		}
	}
	err = s.repository.UpdateModule(&model.Module{ID: moduleID, Status: status})
	if err != nil {
		return model.ModuleStatusUnknown, fmt.Errorf("set module status: %s", err)
	}
	return status, err
}

// SetModuleStatus tries to update status of running module if its possible to rules.
// Returns status of operation, whether status has changed and occurred error
func (s *metadataService) SetOperationStatus(operationID string, status model.OperationStatus, reason string) (model.OperationStatus, bool, error) {
	if operationID == "" {
		return model.OperationStatusUnknown, false, fmt.Errorf("set operation status: no ID provided")
	}
	m, err := s.repository.QueryOperation(&model.Operation{ID: operationID}, 1)
	if err != nil {
		return model.OperationStatusUnknown, false, fmt.Errorf("set operation status: %s", err)
	}
	if len(m) == 0 {
		return model.OperationStatusUnknown, false, fmt.Errorf("set operation status: operation %s not found", operationID)
	}
	if status == m[0].Status {
		return status, false, nil
	}
	// rules to avoid state inconsistency
	switch m[0].Status {
	case model.OperationStatusFailed:
		if status != model.OperationStatusDeleting {
			return model.OperationStatusFailed, false, nil
		}
	case model.OperationStatusDeleting:
		if status != model.OperationStatusDeleteFailed {
			return model.OperationStatusDeleting, false, nil
		}
	case model.OperationStatusDeleteFailed:
		if status != model.OperationStatusDeleting {
			return model.OperationStatusDeleteFailed, false, nil
		}
	}
	err = s.repository.UpdateOperation(&model.Operation{ID: operationID, Status: status, StatusReason: reason})
	if err != nil {
		return model.OperationStatusUnknown, false, fmt.Errorf("set operation status: %s", err)
	}
	return status, true, err
}

// ModuleAfterStart updates module metadata with new information like public ports
func (s *metadataService) ModuleAfterStart(module *model.Module) error {
	if module == nil {
		return nil
	}
	err := s.repository.UpdateModule(&model.Module{
		ID:    module.ID,
		Ports: module.Ports,
	})
	return err
}
