// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package app

import (
	"fmt"
	"github.com/google/uuid"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/logger"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/pkg/mEngine"
	"sync"
)

// service unites all three repository/platform services and provides orchestrator event handling,
// the most critical part of orchestrator.
type orchestratorService struct {
	config 			Config
	// service handling metadata
	metadata 		*metadataService
	// service handling definitions
	definition 		*definitionService
	// service handling platform
	platform 		*platformService
	// logger used in orchestrator
	logger			logger.Logger
	// service grouping and managing correct order of scheduling modules
	mEngine			mEngine.Engine

	// used at teardown to wait for all processes running
	eventJobs 		sync.WaitGroup
	// whether service listens for new event registrations
	openForRegistrations bool
	// channel for events to process in critical part of handling
	criticalChan 		chan Event
	// channel that indicates if teardown of service finished
	doneChan		chan struct{}
}


// Close will handle all buffered requests and gracefully shutdown orchestrator service.
// Not goroutine safe
func (s *orchestratorService) Close() {
	// if Close already called just wait to finish closing
	if !s.openForRegistrations {
		<- s.doneChan
	}
	// stop registrations of new events
	s.openForRegistrations = false
	// wait for already buffered events to finish
	s.eventJobs.Wait()
	// stop critical part loop and start teardown
	close(s.criticalChan)
	// wait for critical part be done
	<- s.doneChan
}

// creates and initialises new Orchestration Service
func newOrchestrationService(meta *metadataService, def *definitionService, pt *platformService, logger logger.Logger, config Config) *orchestratorService {
	const defaultBufferLength = 15

	svc := orchestratorService{
		config: config,
		metadata:        meta,
		definition:      def,
		platform:        pt,
		logger: 		 logger,
		eventJobs:       sync.WaitGroup{},
		openForRegistrations: true,
		doneChan: 		 make(chan struct{}),
		criticalChan: make(chan Event, defaultBufferLength),
		mEngine: mEngine.NewEngine(),
	}
	// start loop for handling of critical part
	go svc.criticalLoop()
	// return service
	return &svc
}

// Registers new event if orchestration service running
func (s *orchestratorService) RegisterEvent(e Event) error {
	var err error
	s.eventJobs.Add(1)
	if !s.openForRegistrations {
		s.eventJobs.Done()
		return fmt.Errorf("register event: service is shut down")
	}
	// prepare and validate event. Return error if validation did not succeeded
	e.register(s)
	err = e.validate()
	if err != nil {
		return err
	}
	err = e.prepare()
	if err != nil {
		return err
	}
	// start event processing
	go func() {
		// unregister event job
		defer s.eventJobs.Done()
		// handle process registered events independently to each other.
		err := e.handle()
		if err != nil {
			return
		}
		// moving to critical part of event processing which is serialized.
		s.criticalChan <- e
	}()
	return nil
}

// criticalLoop represents critical part of event handling, where all events must be processed sequentially
func (s *orchestratorService) criticalLoop() {
	// process all events sequentially until channel closes
	for e := range s.criticalChan  {
		e.critical()
	}
	// stop tasker
	s.mEngine.Close()
	// start teardown of service
	close(s.doneChan)
}

// assignModules runs through every node in operation and try to find the best reusable running modules.
// If reusability is not possible, creates new modules. After run, every node should have module assigned.
func (s *orchestratorService) assignModules(operation *model.Operation, definition *model.OperationDefinition) {
	if operation == nil {
		return
	}
	waitGroup := sync.WaitGroup{}
	type candidate struct {
		// depth of candidate chain; longer chain wins elections as i indicates more modules being reused
		depth int
		// node that proposed candidate to its predecessor
		nodeOfProposer *model.Node
	}
	// all channels to receive candidates from ; to get channel
	candidateReceive := make(map[string][]<-chan []candidate)
	// all channels to send candidates to ; to get channel
	candidateSend := make(map[string][]chan<- []candidate)
	// all channels to send acknowledge of used module
	ackSend := make(map[string][]chan<- *model.Node)
	// all channels to receive acknowledge of used module
	ackReceive := make(map[string][]<-chan *model.Node)
	// create channels between nodes based on their edges
	for _, e := range operation.Edges {
		c := make(chan []candidate, 1)
		a := make(chan *model.Node, 1)

		candidateReceive[e.FromNode] = append(candidateReceive[e.FromNode], c)
		candidateSend[e.ToNode] = append(candidateSend[e.ToNode], c)
		ackReceive[e.ToNode] = append(ackReceive[e.ToNode], a)
		ackSend[e.FromNode] = append(ackSend[e.FromNode], a)
	}

	for index, n := range operation.Nodes {
		index := index
		node := n
		var nodeDef *model.NodeDefinition
		for _, i := range definition.Nodes {
			if i.ID == node.Definition {
				nodeDef = &i
				break
			}
		}
		// get parameters of this particular module
		params := make([]model.ModuleParameter, 0)
		for _, p := range operation.InputParameters {
			if p.ModuleName == nodeDef.ModuleName {
				params = append(params, model.ModuleParameter{
					Key:      p.Key,
					Value:    p.Value,
				})
			}
		}
		waitGroup.Add(1)
		go func() {
			candidates := make(map[string]candidate)
			var finalModuleID string
			var chosen *model.Node

			// setup channels for candidate receive (non root nodes)
			for _, c := range candidateReceive[node.ID] {
				in := c
				cands := <-in
				// for all candidates that I received
				for _, can := range cands {
					// find node predecessors
					predecessor := s.metadata.GetNodePredecessor(can.nodeOfProposer.ID, nodeDef.ModuleName)
					if predecessor == nil {
						continue
					}
					// remove candidate if it is not suitable
					module, err := s.metadata.GetModule(&model.Module{ID: *predecessor.ModuleID}, 1)
					if err != nil || len(module) != 1 {
						continue
					}
					if module[0].Status != model.ModuleStatusRunning ||
						!s.metadata.IsModuleReusable(*module[0], model.Module{Definition: nodeDef.ModuleName, Parameters: params}) {
						continue
					}
					// sort all candidates by used Node to avoid duplications
					can.depth += 1
					candidates[predecessor.ID] = candidate{depth: can.depth + 1, nodeOfProposer: predecessor}
				}
			}
			// find new candidates
			modules, err := s.metadata.GetModule(&model.Module{Definition: nodeDef.ModuleName}, 0)
			if err == nil {
				for _, m := range modules {
					for _, n := range m.InNodes {
						if _, ok := candidates[n.ID]; !ok {
							// check if reusable; must be in running state and have reuasable parameters
							if m.Status == model.ModuleStatusRunning &&
								s.metadata.IsModuleReusable(*m, model.Module{Definition: nodeDef.ModuleName, Parameters: params}) {
								candidates[n.ID] = candidate{depth: 1, nodeOfProposer: &n}
							}
						}
					}
				}
			}
			// send proposals containing candidates - only non root
			for _, out := range candidateSend[node.ID] {
				var proposals []candidate
				for _, c := range candidates {
					proposals = append(proposals, c)
				}
				out <- proposals
				close(out)
			}

			match := true
			// setup channels for acknowledgement receive
			for _, in := range ackReceive[node.ID] {
				ack := <-in
				if ack != nil && match {
					// find node predecessors
					successor := s.metadata.GetNodeSuccessor(ack.ID, nodeDef.ModuleName)
					if successor != nil {
						if _, ok := candidates[successor.ID]; ok {
							if chosen == nil {
								chosen = successor
							}
							if chosen.ID != successor.ID {
								match = false
								chosen = nil
							}
							continue
						}
					}
				}
				match = false
				chosen = nil
			}

			if len(ackReceive[node.ID]) == 0 {
				maxDepth := 0
				for _, c := range candidates {
					if c.depth > maxDepth {
						chosen = c.nodeOfProposer
						finalModuleID = *chosen.ModuleID
					}
				}
			} else if chosen != nil {
				finalModuleID = *chosen.ModuleID
			}
			// code to send own acknowledgements based on decision
			for _, out := range ackSend[node.ID] {
				out <- chosen
				close(out)
			}
			// if not found any module, create own
			if finalModuleID == "" {
				// get module definition
				moduleDef, err := s.definition.GetModule(&model.ModuleDefinition{Name: nodeDef.ModuleName}, 1)
				if err != nil || len(moduleDef) != 1 {
					// todo handle error
					_ = err
				}
				var ports []model.ModulePort
				for _, p := range moduleDef[0].Ports {
					ports = append(ports, model.ModulePort{
						Name:         p.Name,
						Number:       p.Number,
						IsPublic:     p.IsPublic,
					})
				}
				finalModuleID = s.genUUID("module-")
				_, err = s.metadata.AddModule(&model.Module{
					ID:           finalModuleID,
					Status:       model.ModuleStatusScheduling,
					Definition:   nodeDef.ModuleName,
					Parameters:   params,
					Ports:        ports,
				})
				if err != nil {
					// todo handle error
					_ = err
				}
			}
			// put module to node
			operation.Nodes[index].ModuleID = &finalModuleID
			_ = s.metadata.AssignModule(finalModuleID, node.ID)
			waitGroup.Done()
		}()
	}
	waitGroup.Wait()
}

// getModuleStatus returns status of module based on information in repository and also from platform.
func (s *orchestratorService) getModuleStatus(moduleID string) model.ModuleStatus {
	var dbStatus, ptStatus model.ModuleStatus
	var err error

	m, err := s.metadata.GetModule(&model.Module{ID: moduleID}, 1)
	if err != nil || len(m) == 0 {
		return model.ModuleStatusUnknown
	}
	dbStatus = m[0].Status
	if dbStatus == model.ModuleStatusCompleted || dbStatus == model.ModuleStatusFailed {
		return dbStatus
	}
	if ptStatus, err = s.platform.StatusModule(moduleID); err != nil {
		ptStatus = model.ModuleStatusUnknown
	}
	if (dbStatus == model.ModuleStatusCreating || dbStatus == model.ModuleStatusScheduling) && ptStatus == model.ModuleStatusUnknown {
		return dbStatus
	}
	_,_ = s.metadata.SetModuleStatus(moduleID, ptStatus)
	return ptStatus
}

// isModuleReady return true if module is running or completed
func (s *orchestratorService) isModuleReady(module *model.Module) bool {
	status := s.getModuleStatus(module.ID)
	// todo also possible failed state
	if status == model.ModuleStatusRunning || status == model.ModuleStatusCompleted {
		return true
	}
	return false
}

// generates new uuid
func (s *orchestratorService) genUUID(prefix string) string {
	return prefix + uuid.New().String()
}


func (s *orchestratorService) modulesFromSubtree(operation *model.Operation, nodeID string) []model.Module {
	// todo error handle
	foundNodes := make(map[string]struct{})
	result := make([]model.Module, 0)
	var getParamsRecursive func(node *model.Node)
	getParamsRecursive = func(node *model.Node) {
		if _, ok := foundNodes[node.ID]; ok {
			return
		} else {
			foundNodes[nodeID] = struct{}{}
		}
		module, _ := s.metadata.GetModule(&model.Module{ID: *node.ModuleID}, 0)
		result = append(result, *module[0])
		for _, previous := range s.previousNodes(operation, node.ID) {
			getParamsRecursive(previous)
		}
	}
	// testing
	n, err := s.metadata.repository.QueryNode(&model.Node{ID: nodeID}, 1)
	if err != nil || len(n) == 0 {
		return result
	}
	getParamsRecursive(n[0])
	//for _, previous := range s.previousNodes(operation, nodeID) {
	//	getParamsRecursive(previous)
	//}
	return result
}

// takes operation and nodeID and returns nodes that are predecessors to node with NodeID
func (s *orchestratorService) previousNodes(operation *model.Operation, nodeID string) []*model.Node {
	nodes := make([]*model.Node, 0)
	edges, err := s.metadata.repository.QueryEdge(&model.Edge{
		OperationID: operation.ID,
		ToNode:      nodeID,
	}, 0)
	if err != nil {
		return nodes
	}
	for _, e := range edges {
		n, err := s.metadata.repository.QueryNode(&model.Node{ID: e.FromNode}, 1)
		if err != nil || len(n) != 1 {
			continue
		}
		nodes = append(nodes, n[0])
	}
	return nodes
}