// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package app

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/chi-middleware/logrus-logger"
	"github.com/go-chi/chi"
	"github.com/go-chi/chi/middleware"
	"github.com/go-chi/cors"
	"github.com/go-chi/render"
	"github.com/jinzhu/gorm"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/pkg/bodyDecode"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/pkg/respond"
	"log"
	"net/http"
)

type apiHandler struct {
	config Config
	definition *definitionService
	metadata *metadataService
	orchestrator *orchestratorService
}

func (o *Orchestrator) GetRoutes() http.Handler {
	handler := apiHandler{o.config, o.definition, o.metadata, o.service}
	h := chi.NewRouter()
	h.Use(
		cors.Handler(cors.Options{
			AllowedMethods:   []string{"GET", "POST", "PUT", "DELETE"},
			AllowedHeaders:   []string{"Accept", "Authorization", "Content-Type", "X-CSRF-Token", "X-PINGOTHER"},
			AllowCredentials: false,
			ExposedHeaders:   []string{"Link"},
			MaxAge:           300, // Maximum value not ignored by any of major browsers
		}),
		render.SetContentType(render.ContentTypeJSON),
		logger.Logger("router", o.logger.GetLogger()),
		middleware.RedirectSlashes,
		middleware.Recoverer,
	)
	h.Route("/v1", func(r chi.Router) {
		r.Get("/config", handler.getConfig)
		r.Route("/definition", handler.getDefinitionRoutes)
		r.Route("/module", handler.getModuleRoutes)
		r.Route("/operation", handler.getOperationRoutes)
		r.Route("/node", handler.getNodeRoutes)
	})
	return h
}

func (h *apiHandler) getDefinitionRoutes(r chi.Router) {
	r.Get("/operation", h.listOperationDefinition)
	r.Get("/operation/{operationName}", h.getOperationDefinition)
	r.Get("/node/{operationName}",h.listNodeDefinitionByOperation)
	r.Get("/module", h.listModuleDefinition)
	r.Get("/module/{moduleName}", h.getModuleDefinition)
}

func (h *apiHandler) getOperationRoutes(r chi.Router) {
	r.Get("/", h.listRunningOperation)
	r.Post("/", h.createRunningOperation)
	r.With(h.operationCtx).Get("/{operationID}", h.getRunningOperation)
	r.With(h.operationCtx).Delete("/{operationID}", h.deleteRunningOperation)
}

func (h *apiHandler) getNodeRoutes(r chi.Router) {
	r.Get("/{operationID}", h.listRunningNodeByOperation)
}

func (h *apiHandler) getModuleRoutes(r chi.Router) {
	r.Get("/", h.listRunningModule)
	r.With(h.moduleCtx).Get("/{moduleID}", h.getRunningModule)
}

func (h *apiHandler) operationCtx(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		operationID := chi.URLParam(r, "operationID")
		ops, err := h.metadata.GetOperation(&model.Operation{ID: operationID}, 1)
		if len(ops) == 0 {
			http.Error(w, http.StatusText(http.StatusNotFound), http.StatusNotFound)
			return
		} else if err != nil {
			http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
			return
		}
		ctx := context.WithValue(r.Context(), "operation", ops[0])
		next.ServeHTTP(w, r.WithContext(ctx))
	})
}

func (h *apiHandler) moduleCtx(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		moduleID := chi.URLParam(r, "moduleID")
		module, err := h.metadata.GetModule(&model.Module{ID: moduleID}, 1)
		if len(module) == 0 {
			http.Error(w, http.StatusText(http.StatusNotFound), http.StatusNotFound)
			return
		} else if err != nil {
			http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		}
		module[0].Status = h.orchestrator.getModuleStatus(moduleID)
		ctx := context.WithValue(r.Context(), "module", module[0])
		next.ServeHTTP(w, r.WithContext(ctx))
	})
}

// Get configuration of orchestrator
func (h *apiHandler) getConfig(w http.ResponseWriter, _ *http.Request) {
	type config struct {
		DataWarehouse string
	}
	out := config{
		DataWarehouse: h.config.Warehouse,
	}
	resp, err := json.Marshal(&out)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
}

func (h *apiHandler) getOperationDefinition(w http.ResponseWriter, r *http.Request) {
	name := chi.URLParam(r, "operationName")
	defs, err := h.definition.GetOperation(&model.OperationDefinition{Name: name}, 1)
	if len(defs) == 0 {
		http.Error(w, http.StatusText(http.StatusNotFound), http.StatusNotFound)
		return
	} else if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}

	resp, err := json.MarshalIndent(defs[0], "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listOperationDefinition(w http.ResponseWriter, _ *http.Request) {
	defs, err := h.definition.GetOperation(nil, 0)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(defs, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) getModuleDefinition(w http.ResponseWriter, r *http.Request) {
	name := chi.URLParam(r, "moduleName")
	defs, err := h.definition.GetModule(&model.ModuleDefinition{Name: name}, 1)
	if len(defs) == 0 {
		http.Error(w, http.StatusText(http.StatusNotFound), http.StatusNotFound)
		return
	} else if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.Marshal(defs[0])
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listModuleDefinition(w http.ResponseWriter, _ *http.Request) {
	defs, err := h.definition.GetModule(nil, 0)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
	}
	resp, err := json.MarshalIndent(&defs, "", "   ")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listNodeDefinitionByOperation(w http.ResponseWriter, r *http.Request) {
	opName := chi.URLParam(r, "operationName")
	defs, err := h.definition.GetNode(&model.NodeDefinition{OperationName: opName},0)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(&defs, "", "   ")
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) getRunningNode(w http.ResponseWriter, r *http.Request) {
	ctx := r.Context()
	node, ok := ctx.Value("node").(*model.Node)
	if !ok {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(&node, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listRunningNodeByOperation(w http.ResponseWriter, r *http.Request) {
	opID := chi.URLParam(r, "operationID")
	nodes, err := h.metadata.GetNode(&model.Node{OperationID: opID}, 0)
	if gorm.IsRecordNotFoundError(err) {
		http.Error(w, http.StatusText(http.StatusNotFound), http.StatusNotFound)
		return
	} else if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(&nodes, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) getRunningOperation(w http.ResponseWriter, r *http.Request) {
	ctx := r.Context()
	op, ok := ctx.Value("operation").(*model.Operation)
	if !ok {
		http.Error(w, http.StatusText(http.StatusUnprocessableEntity), http.StatusUnprocessableEntity)
		return
	}
	resp, err := json.MarshalIndent(&op, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listRunningOperation(w http.ResponseWriter, _ *http.Request) {
	ops, err := h.metadata.GetOperation(nil, 0)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(&ops, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) getRunningModule(w http.ResponseWriter, r *http.Request) {
	ctx := r.Context()
	node, ok := ctx.Value("module").(*model.Module)
	if !ok {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	resp, err := json.MarshalIndent(&node, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) listRunningModule(w http.ResponseWriter, _ *http.Request) {
	ops, err := h.metadata.GetModule(nil, 0)
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	for i, o := range ops {
		ops[i].Status = h.orchestrator.getModuleStatus(o.ID)
	}
	resp, err := json.MarshalIndent(&ops, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) createRunningOperation(w http.ResponseWriter, r *http.Request) {
	type OpCreate struct {
		ID string
		Parameters map[string]map[string]string
		Name string
	}
	var request OpCreate
	// transform body to request data
	err := bodyDecode.JSON(w, r, &request)
	if err != nil {
		var mr *bodyDecode.MalformedRequest
		if errors.As(err, &mr) {
			http.Error(w, mr.Msg, mr.Status)
		} else {
			log.Println(err.Error())
			http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		}
		return
	}
	// create and register new start operation event
	event := &StartOperationEvent{ID: request.ID, Definition: request.Name, Parameters: request.Parameters}
	err = h.orchestrator.RegisterEvent(event)
	if err != nil {
		var badRequestError *ErrorBadRequest
		var internalError *ErrorInternal
		if errors.As(err, &badRequestError) {
			http.Error(w, badRequestError.Msg, http.StatusBadRequest)
			return
		}
		if errors.As(err, &internalError) {
			http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
			return
		}
	}
	response := struct {
		ID string
	}{
		ID: event.response,
	}
	resp, err := json.MarshalIndent(&response, "", "   ")
	if err != nil {
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
		return
	}
	respond.WithJSON(w, http.StatusOK, resp)
	return
}

func (h *apiHandler) deleteRunningOperation(w http.ResponseWriter, r *http.Request) {
	ctx := r.Context()
	op, ok := ctx.Value("operation").(*model.Operation)
	if !ok {
		http.Error(w, http.StatusText(http.StatusUnprocessableEntity), http.StatusUnprocessableEntity)
		return
	}
	event := &DeleteOperationEvent{OperationID: op.ID}
	_ = h.orchestrator.RegisterEvent(event)
	respond.WithString(w, http.StatusOK, fmt.Sprintf("operation %s scheduled to delete", op.ID))
	return
}

