// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package config

import (
	"github.com/spf13/viper"
)

// GetConfig reads all orchestrator configuration options from provided file.
// In case of missing option, it sets default value.
func GetConfig(path string) (Config, error) {
	// set options
	viper.SetConfigType("yaml")
	if path != "" {
		viper.SetConfigFile(path)
	} else {
		viper.SetConfigName("config")
		viper.AddConfigPath(".")
		viper.AddConfigPath("config/")
		viper.AddConfigPath("$HOME/.orchestrator")
		viper.AddConfigPath("/etc/orchestrator/")
	}
	// try to parse configuration file, return in case of error
	err := viper.ReadInConfig()
	if err != nil {
		return Config{}, err
	}

	// call dedicated config functions on each section
	definitionConf := getDB(viper.Sub("definitions"))
	metadataConf := getDB(viper.Sub("metadata"))
	platformConf := getPlatform(viper.Sub("platform"))
	apiConf := getApi(viper.Sub("api"))
	logConf := getLog(viper.Sub("log"))
	dockerConf := getDocker(viper.Sub("docker"))
	warehouseConf := getWarehouse(viper.Sub("warehouse"))
	// set sections into config struct
	c := Config{
		Definitions: definitionConf,
		Metadata:    metadataConf,
		Platform:    platformConf,
		Log:         logConf,
		Warehouse:   warehouseConf,
		Api:         apiConf,
		Docker: 	 dockerConf,
	}
	// return config
	return c, nil
}

func getApi(v *viper.Viper) api {
	const (
		// all addresses
		defaultAddress = ""
		defaultPort    = "8080"
	)
	var (
		address string
		port    string
	)

	if s := v.GetString("listen"); s == "" {
		address = defaultAddress
	} else {
		address = s
	}

	if s := v.GetString("port"); s == "" {
		port = defaultPort
	} else {
		port = s
	}

	return api{
		Address: address,
		Port:    port,
	}
}

func getDocker(v *viper.Viper) docker {
	var t string
	if v != nil {
		t = v.GetString("registry")
	}
	c := docker{Registry: t}
	return c
}

func getPlatform(v *viper.Viper) platform {
	// default values handled in platform package
	t := v.GetString("type")
	c := platform{
		PlatformType: t,
		Options:      v.GetStringMap(t),
	}
	return c
}

func getDB(v *viper.Viper) db {
	// default values handled in database package
	opts := dbOptions{}
	t := v.GetString("type")
	if v.Sub(t) != nil {
		opts.Host = v.Sub(t).GetString("host")
		opts.Port = v.Sub(t).GetString("port")
		opts.Database = v.Sub(t).GetString("database")
		opts.User = v.Sub(t).GetString("username")
		opts.Password = v.Sub(t).GetString("password")
	}
	c := db{
		DBType:  t,
		Options: opts,
	}
	return c
}

func getLog(v *viper.Viper) log {
	const (
		defaultLogFile = "/tmp/orchestrator/orchestrator.log"
		defaultDebug   = false
	)
	logfile := defaultLogFile
	if s := v.GetString("logfile"); s != "" {
		logfile = s
	}
	debug := defaultDebug
	if d := v.GetBool("debug"); d {
		debug = d
	}
	return log{
		Logfile: logfile,
		Debug:   debug,
	}
}

func getWarehouse(v *viper.Viper) warehouse {
	const (
		defaultHost = "127.0.0.1"
		defaultPort = "8081"
	)
	host := defaultHost
	port := defaultPort
	if s := v.GetString("host"); s != "" {
		host = s
	}
	if s := v.GetString("port"); s != "" {
		port = s
	}
	return warehouse{
		Host: host,
		Port: port,
	}
}
