// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package model

import (
	"fmt"
	"github.com/jinzhu/gorm"
)

// OperationDefinition represents definition of operation, which is used to determine the workflow of running operation.
type OperationDefinition struct {
	// operation name
	Name string `gorm:"primary_key"`
	// description for user
	Description string
	// nodes that operation workflow consists of
	Nodes []NodeDefinition `gorm:"ForeignKey:OperationName;AssociationForeignKey:Name"`
	// edges of operation workflow, represent connections between nodes
	Edges []EdgeDefinition `gorm:"ForeignKey:OperationName;AssociationForeignKey:Name"`
}

// Returns name of table in database that stores OperationDefinition models.
func (OperationDefinition) TableName() string {
	return "definition_operations"
}

// AfterDelete hook defined for cascade delete when gorm is used
func (d *OperationDefinition) AfterDelete(tx *gorm.DB) error {
	err := tx.Model(&NodeDefinition{}).Unscoped().Delete(&d.Nodes).Error
	err = tx.Model(&EdgeDefinition{}).Unscoped().Delete(&d.Edges).Error
	return err
}

// NodeDefinition represents one node definition of operation workflow. Each node holds a module
// which should be run when node is invoked. It holds overrides of module parameters.
type NodeDefinition struct {
	// unique identifier of node definition
	ID uint `gorm:"primary_key"`
	// name of module that should start when node invoked
	// each node in operation must hold its own unique module
	ModuleName string
	// description for user
	Description string
	// operation that the node is part of
	OperationName string `gorm:"primary_key"`
	// Specifies overrides for underlying module parameters
	Parameters []NodeDefinitionParameter `gorm:"ForeignKey:NodeID;AssociationForeignKey:ID"`
}

// Returns name of table in database that stores NodeDefinition models.
func (NodeDefinition) TableName() string {
	return "definition_nodes"
}

// AfterDelete hook defined for cascade delete when gorm used
func (d *NodeDefinition) AfterDelete(tx *gorm.DB) error {
	err := tx.Model(&NodeDefinitionParameter{}).Unscoped().Delete(&d.Parameters).Error
	return err
}

// NodeDefinitionParameter represent overrides for underlying module parameter properties at node
type NodeDefinitionParameter struct {
	// node name that parameter belongs to
	NodeID uint `gorm:"primary_key;auto_increment:false"`
	// parameter name
	Name string `gorm:"primary_key"`
	// default parameter value; optional
	DefaultValue *string
	// whether the parameter can be overridden by input parameters specified at request
	IsConstant bool
}

// Returns name of table in database that stores NodeDefinitionParameter models.
func (NodeDefinitionParameter) TableName() string {
	return "definition_node_parameters"
}

// EdgeDefinition represents a relation between two workflow nodes. This determines order of node execution
// when start of operation is invoked.
type EdgeDefinition struct {
	// name of operation this edge belongs to
	OperationName string `gorm:"primary_key"`
	// identifier of node that must run firstly
	FromNode uint `gorm:"primary_key;auto_increment:false"`
	// identifier of node that must run secondly
	ToNode uint `gorm:"primary_key;auto_increment:false"`
}

// Returns name of table in database that stores EdgeDefinition models.
func (EdgeDefinition) TableName() string {
	return "definition_edges"
}

// ModuleDefinition represents module definition. Holds image name that should run when module invoked.
// Also holds default parameters and ports for module.
type ModuleDefinition struct {
	// module name
	Name string `gorm:"primary_key"`
	// type of module
	Type ModuleType
	// description for user
	Description string
	// whether the running of module blocks next modules
	IsBlocking bool
	// port used for orchestrator to see if module is ready; optional
	HealthPort int
	// name of docker image used
	Image string
	// docker image tag
	Tag string
	// module parameters
	Parameters []ModuleDefinitionParameter `gorm:"foreignkey:Module;association_foreignkey:Name"`
	// list of all ports that module use
	Ports []ModuleDefinitionPort `gorm:"foreignkey:Module;association_foreignkey:Name"`
	// resource constraints for module
	Resources ModuleDefinitionResources `gorm:"foreignkey:Module;association_foreignkey:Name"`
}

// Represents type of module
type ModuleType string
const (
	// module of type job is running a program that should finish at some time. Typically next modules wait for it.
	ModuleTypeJob     ModuleType = "job"
	// module of type service is long running service that runs until whole operation is deleted.
	ModuleTypeService ModuleType = "service"
)

// Returns name of table in database that stores ModuleDefinition models.
func (ModuleDefinition) TableName() string {
	return "definition_modules"
}

// BeforeSave hook defined to check module type when gorm used
func (d *ModuleDefinition) BeforeSave() (err error) {
	if  d.Type != ModuleTypeService && d.Type != ModuleTypeJob {
		return fmt.Errorf("cannot save module definition, unsupported type: %s", d.Type)
	}
	return
}

// AfterDelete hook defined for cascade delete when gorm used
func (d *ModuleDefinition) AfterDelete(tx *gorm.DB) error {
	err := tx.Model(&ModuleDefinitionParameter{}).Unscoped().Delete(&d.Parameters).Error
	err = tx.Model(&ModuleDefinitionPort{}).Unscoped().Delete(&d.Ports).Error
	err = tx.Model(&ModuleDefinitionResources{}).Unscoped().Delete(&d.Resources).Error
	return err
}

// ModuleDefinitionParameter represents parameter of module
type ModuleDefinitionParameter struct {
	// name of operation parameter belongs to
	Module string `gorm:"primary_key"`
	// parameter name
	Name string `gorm:"primary_key"`
	// description of parameter
	Description string
	// default parameter value (used when param not specified)
	DefaultValue string
	// if set to false, parameter can be different at dependency check
	IsStrict bool
}

// Returns name of table in database that stores ModuleDefinitionParameter models.
func (ModuleDefinitionParameter) TableName() string {
	return "definition_module_parameters"
}

// ModuleDefinitionPort represents port of module
type ModuleDefinitionPort struct {
	// name of the module that port belongs to
	Module string `gorm:"primary_key"`
	// port name
	Name string `gorm:"primary_key"`
	// port number
	Number int
	// whether the port should be open to public network or not
	IsPublic bool
}

// Returns name of table in database that stores ModuleDefinitionPort models.
func (ModuleDefinitionPort) TableName() string {
	return "definition_module_ports"
}

// ModuleDefinitionResources represents resource constraints of module
type ModuleDefinitionResources struct {
	// name of the module these resource constraints belongs to
	Module string `gorm:"primary_key"`
	// number of cpu cores that module needs
	Cpus int
	// minimal amount of ram in megabytes that module needs
	Ram int
}

// Returns name of table in database that stores ModuleDefinitionPort models.
func (ModuleDefinitionResources) TableName() string {
	return "definition_module_resources"
}

//// BeforeCreate hook defined to check if port number not already defined when gorm used
//func (u *ModuleDefinitionPort) BeforeCreate(tx *gorm.DB) error {
//	var s ModuleDefinitionPort
//	if err := tx.Model(&ModuleDefinitionPort{}).
//		Where("module = ? AND number = ?", u.Module, u.Number).
//		First(&s).Error; gorm.IsRecordNotFoundError(err) {
//			return nil
//	} else if err != nil {
//		return err
//	}
//	return fmt.Errorf("cannot save module port: port numer %d already associated with module %s",
//			u.Number, u.Module)
//}

//// BeforeUpdate hook defined port number not already defined when gorm used
//func (u *ModuleDefinitionPort) BeforeUpdate(tx *gorm.DB) error {
//	var s ModuleDefinitionPort
//	if err := tx.Model(&ModuleDefinitionPort{}).
//		Where("module = ? AND number = ?", u.Module, u.Number).
//		First(&s).Error; err == nil && u.Name != s.Name {
//		return fmt.Errorf("cannot update module port: port numer %d already associated with module %s",
//			u.Number, u.Module)
//	} else if err != nil {
//		return err
//	}
//	return nil
//}