// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package model

import (
	"bytes"
	"github.com/jinzhu/gorm"
)

// Operation represents model of running operation (metadata).
type Operation struct {
	// ID of running operation
	ID string `gorm:"primary_key"`
	// name of operation definition used
	Definition string
	// all input arguments specified at the api call for the operation
	InputParameters []OperationParameter `gorm:"ForeignKey:OperationID;AssociationForeignKey:ID"`
	// status/phase of operation
	Status OperationStatus
	// Reason of status (used as description when status failed)
	StatusReason string
	// running nodes associated with this operation
	Nodes []Node `gorm:"ForeignKey:OperationID;AssociationForeignKey:ID"`
	// connections between nodes
	Edges []Edge `gorm:"ForeignKey:OperationID;AssociationForeignKey:ID"`
}

// OperationStatus represents status (phase) of running operation.
type OperationStatus int
const (
	// Status of operation at creation process before platform any of nodes have module created
	OperationStatusScheduling OperationStatus = iota
	// Status of operation when at least one module is already running
	OperationStatusRunning
	// Status of operation when error happened at the run of any module assigned.
	OperationStatusFailed
	// Status of operation when user triggered the deletion process of operation
	OperationStatusDeleting
	// Status of operation when delete has failed
	OperationStatusDeleteFailed
	// Status of operation when user triggered the deletion process of operation
	// Used when status of operation could not be determined
	OperationStatusUnknown
)

func (s OperationStatus) String() string {
	switch s {
	case OperationStatusScheduling:
		return "Scheduling"
	case OperationStatusRunning:
		return "Running"
	case OperationStatusFailed:
		return "Failed"
	case OperationStatusDeleting:
		return "Deleting"
	case OperationStatusDeleteFailed:
		return "Delete Failed"
	}
	return ""
}

// MarshalJSON marshals the enum as a quoted json string
func (s OperationStatus) MarshalJSON() ([]byte, error) {
	buffer := bytes.NewBufferString(`"`)
	buffer.WriteString(s.String())
	buffer.WriteString(`"`)
	return buffer.Bytes(), nil
}

// Returns name of table in database that stores Operation models
func (Operation) TableName() string {
	return "metadata_operations"
}

// AfterDelete hook defined for cascade delete when using gorm
func (o *Operation) AfterDelete(tx *gorm.DB) error {
	err := tx.Model(&OperationParameter{}).Unscoped().Delete(&o.InputParameters).Error
	err = tx.Model(&Node{}).Unscoped().Delete(&o.Nodes).Error
	err = tx.Model(&Edge{}).Unscoped().Delete(&o.Edges).Error
	return err
}

// OperationParameter represents model of input parameter used at creation of operation
type OperationParameter struct {
	// id of running operation (deployment name)
	OperationID string `gorm:"primary_key"`
	// name of module that parameter belongs to
	ModuleName string `gorm:"primary_key"`
	// parameter name
	Key string `gorm:"primary_key"`
	// parameter value
	Value string `gorm:"not null"`
}

// Returns name of table in database that stores OperationParameter models
func (OperationParameter) TableName() string {
	return "metadata_operation_parameters"
}

// Node represents model of running node assigned to operation. It holds running module which could be created 
// or reused from other running nodes based if checks are satisfied 
type Node struct {
	// ID of running node
	ID string `gorm:"primary_key"`
	// ID of operation node belongs to
	OperationID string
	// Reference to running module that Node holds
	ModuleID *string
	// ID of node definition 
	Definition uint
}

// Returns name of table in database that stores Node models
func (Node) TableName() string {
	return "metadata_nodes"
}

// AfterDelete hook defined for cascade delete when gorm used
func (m *Node) AfterDelete(tx *gorm.DB) error {
	if m.ModuleID != nil && *m.ModuleID != ""  {
		var module Module
		if tx.Model(&Module{}).Where(&Module{ID: *m.ModuleID}).First(&module).RecordNotFound() {
			return nil
		} else if len(module.InNodes) == 0 {
			return tx.Model(&Module{}).Unscoped().Delete(&module).Error
		}
	}
	return nil
}

// Edge represents connection between two running nodes. It has direction which is described by attributes "FromNode"
// and "ToNode". Node described by "FromNode" is parent to Node described by "ToNode". 
// In workflow "FromNode" must always run before "ToNode".
type Edge struct {
	// ID of operation edge belongs to
	OperationID string
	// ID of running node that must run first
	FromNode string `gorm:"primary_key"`
	// ID of running node that must run second
	ToNode string `gorm:"primary_key"`
}

// Returns name of table in database that stores Edge models
func (Edge) TableName() string {
	return "metadata_edges"
}

// Module represents running module. It must belong to at least one Node. Should get deleted when last node is removed.
type Module struct {
	// ID of running module
	ID string `gorm:"primary_key"`
	// status of module
	Status ModuleStatus
	// Reason of status (used as description when status failed)
	StatusReason string
	// mame of module definition used
	Definition string
	// References to all workflow nodes that use this module.
	// Scheduler uses this attribute to check if module is going to delete soon.
	InNodes []Node `gorm:"ForeignKey:ModuleID;AssociationForeignKey:ID"`
	// all input arguments specified at creation of module
	Parameters []ModuleParameter `gorm:"ForeignKey:ModuleID;AssociationForeignKey:ID"`
	// All ports that module has open
	Ports []ModulePort `gorm:"ForeignKey:ModuleID;AssociationForeignKey:ID"`
}

// ModuleStatus represents status of running module.
type ModuleStatus int
const (
	// Status of module at the creation process before actual run on platform.
	ModuleStatusCreating ModuleStatus = iota
	//
	ModuleStatusScheduling
	// Status of module when created and running on platform
	ModuleStatusRunning
	// Status of module when any error occurred at the process of creation, scheduling or run.
	ModuleStatusFailed
	// Status of module when completed without any errors. Can be used only with module type "Job"
	ModuleStatusCompleted
	// Used when status of module could not be determined
	ModuleStatusUnknown
	// Status of module when completed without any errors. Can be used only with module type "Job"
	ModuleStatusDeleting
	// Status of module when completed without any errors. Can be used only with module type "Job"
	ModuleStatusDeleteFailed
)

func (s ModuleStatus) String() string {
	switch s {
	case ModuleStatusCreating:
		return "Creating"
	case ModuleStatusRunning:
		return "Running"
	case ModuleStatusFailed:
		return "Failed"
	case ModuleStatusCompleted:
		return "Completed"
	case ModuleStatusDeleting:
		return "Deleting"
	case ModuleStatusDeleteFailed:
		return "Delete Failed"
	case ModuleStatusScheduling:
		return "Scheduling"
	}
	return ""
}

// MarshalJSON marshals the enum as a quoted json string
func (s ModuleStatus) MarshalJSON() ([]byte, error) {
	buffer := bytes.NewBufferString(`"`)
	buffer.WriteString(s.String())
	buffer.WriteString(`"`)
	return buffer.Bytes(), nil
}

// Returns name of table in database that stores Module models.
func (Module) TableName() string {
	return "metadata_modules"
}

// AfterDelete hook defined for cascade delete when gorm used
func (m *Module) AfterDelete(tx *gorm.DB) error {
	var err error
	err = tx.Model(&ModulePort{}).Unscoped().Delete(&m.Ports).Error
	err = tx.Model(&ModuleParameter{}).Unscoped().Delete(&m.Parameters).Error
	return err
}

// ModuleParameter represents parameter of running module.
type ModuleParameter struct {
	// name of running module (deployment name)
	ModuleID string `gorm:"primary_key"`
	// name of module that parameter belongs to
	//ModuleName string `gorm:"primary_key"`
	// parameter name
	Key string `gorm:"primary_key"`
	// parameter value
	Value string `gorm:"not null"`
}

// Returns name of table in database that stores ModuleParameter models.
func (ModuleParameter) TableName() string {
	return "metadata_module_parameters"
}

// ModulePort represents open port of running module
type ModulePort struct {
	// name of running module (deployment name)
	ModuleID string `gorm:"primary_key"`
	// name of port (human readable identifier); each port of module must have unique name
	Name string `gorm:"primary_key"`
	// port number used in network internal to platform (generally only between running modules)
	Number int `gorm:"not null"`
	// address with port used in public network when chosen to be exposed
	PublicEndpoint string `gorm:"not null"`
	// whether or not this port should be exposed to external network
	IsPublic bool
}

// Returns name of table in database that stores ModulePort models
func (ModulePort) TableName() string {
	return "metadata_module_ports"
}
