// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package model

// DefinitionRepository provides methods to manipulate with definition data in storage.
type DefinitionStorer interface {
	// Stores specified operation definitions in repository.
	// Returns ErrAlreadyStored when at least one of definitions already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreOperation(...*OperationDefinition) error

	// Stores specified node definitions in repository.
	// Returns ErrInvalidParameter if no operation with name "OperationName" or
	// no module with name "ModuleName" already stored.
	// Returns ErrAlreadyStored when at least one of definitions already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreNode(...*NodeDefinition) error

	// Store specified edge definitions in repository.
	// Returns ErrInvalidParameter if no operation with name "OperationName" or
	// no nodes with names "FromNode" and "ToNode" already stored.
	// Returns ErrAlreadyStored when at least one of definitions already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreEdge(...*EdgeDefinition) error

	// Stores specified module definitions in repository.
	// Returns ErrInvalidParameter when specified Type not supported.
	// Returns ErrAlreadyStored when at least one of definitions already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreModule(...*ModuleDefinition) error

	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateOperation(*OperationDefinition) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateNode(*NodeDefinition) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateEdge(*EdgeDefinition) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateModule(*ModuleDefinition) error

	// Deletes specified operations from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteOperation(...*OperationDefinition) error
	// Deletes specified modules from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteModule(...*ModuleDefinition) error
	// Deletes specified nodes from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteNode(...*NodeDefinition) error
	// Deletes specified edges from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteEdge(...*EdgeDefinition) error
}

// DefinitionFetcher provides methods to fetch definition data from storage.
type DefinitionFetcher interface {
	// Fetch number of operations from store that satisfies specified query
	// If query = nil, all stored operations get fetched
	// Parameter limit specifies maximum number of operations that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryOperation(query *OperationDefinition, limit int) ([]*OperationDefinition, error)

	// Fetch number of nodes from store that satisfies specified query
	// If query = nil, all stored nodes get fetched
	// Parameter limit specifies maximum number of nodes that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryNode(query *NodeDefinition, limit int) ([]*NodeDefinition, error)

	// Fetch number of edges from store that satisfies specified query
	// If query = nil, all stored edges get fetched
	// Parameter limit specifies maximum number of edges that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryEdge(query *EdgeDefinition, limit int) ([]*EdgeDefinition, error)

	// Fetch number of modules from store that satisfies specified query
	// If query = nil, all stored modules get fetched
	// Parameter limit specifies maximum number of modules that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryModule(query *ModuleDefinition, limit int) ([]*ModuleDefinition, error)
}

// MetadataManager provides methods to manipulate with metadata structures in storage.
type MetadataManager interface{
	// Stores specified operations in repository.
	// Returns ErrAlreadyStored when at least one of operations with specified ID already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreOperation(*Operation) (*Operation, error)
	// Stores specified nodes in repository.
	// Returns ErrInvalidParameter if no operation with ID "OperationID" or
	// no module with name "ModuleName" already stored.
	// Returns ErrAlreadyStored when at least one of nodes with specified ID already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreNode(*Node) (*Node, error)
	// Store specified edges in repository.
	// Returns ErrInvalidParameter if no operation with ID "OperationID" or
	// no nodes with IDs "FromNode" and "ToNode" already stored.
	// Returns ErrAlreadyStored when at least one of edges already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreEdge(*Edge) (*Edge, error)
	// Stores specified modules in repository.
	// Returns ErrInvalidParameter when specified Type not supported.
	// Returns ErrAlreadyStored when at least one of modules with specified ID already stored.
	// Returns ErrInternal when connection lost or repository has internal issues.
	StoreModule(*Module) (*Module, error)

	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateOperation(*Operation) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateNode(*Node) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateEdge(*Edge) error
	// Updates specified operation definition in repository.
	// Returns ErrNotFound when operation with specified was not found.
	// Returns ErrInternal when connection lost or repository has internal issues.
	UpdateModule(*Module) error

	// Deletes specified operations from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteOperation(...*Operation) error
	// Deletes specified nodes from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteNode(...*Node) error
	// Deletes specified edges from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteEdge(...*Edge) error
	// Deletes specified modules from repository.
	// Returns ErrInternal when connection lost or repository has internal issues.
	DeleteModule(...*Module) error

	// Fetch number of operations from repository that satisfies specified query
	// If query = nil, all stored operations get fetched
	// Parameter limit specifies maximum number of operations that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryOperation(query *Operation, limit int) ([]*Operation, error)

	// Fetch number of nodes from repository that satisfies specified query
	// If query = nil, all stored nodes get fetched
	// Parameter limit specifies maximum number of nodes that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryNode(query *Node, limit int) ([]*Node, error)

	// Fetch number of edges from repository that satisfies specified query
	// If query = nil, all stored edges get fetched
	// Parameter limit specifies maximum number of edges that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryEdge(query *Edge, limit int) ([]*Edge, error)

	// Fetch number of modules from repository that satisfies specified query
	// If query = nil, all stored modules get fetched
	// Parameter limit specifies maximum number of modules that get fetched.
	// Returns ErrInternal when connection lost or repository has internal issues.
	QueryModule(query *Module, limit int) ([]*Module, error)
}