// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package kubernetes

import (
	"fmt"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/clientcmd"
	"sync"
)

const (
	defaultNamespace = "default"
	defaultKubeConfig = "config/kube.config"
	defaultPublicBindType = "NodePort"
	defaultPublicAddress = "127.0.0.1"
)

// Configuration options for kubernetes manager
type Config struct {
	// Type of mechanism to use when creating external services
	PublicBindType string
	// Address of public network, which is used for endpoint creation
	PublicAddress string
	// Path to kubeConfig file; valid only for use when outside cluster
	KubeConfig string
	// kubernetes namespace to use for modules
	Namespace string
	// whether or not is orchestrator running inside (kubernetes) cluster
	InCluster bool
}

// Implementation of PlatformManager using Kubernetes as a driver
type kubernetesManager struct {
	c       kubernetes.Interface
	config  Config
	sync    sync.Mutex
}

// ParseConfig reads configuration options from provided map and returns parsed Config struct
func ParseConfig(opts map[string]interface{}) Config {
	config := Config{
		PublicBindType: defaultPublicBindType,
		PublicAddress: defaultPublicAddress,
		KubeConfig: defaultKubeConfig,
		Namespace: defaultNamespace,
	}
	if opt, ok := opts["public_address_bind_type"]; ok {
		if c, ok := opt.(string); ok {
			config.PublicBindType = c
		}
	}
	if opt, ok := opts["nodeport_address"]; ok {
		if c, ok := opt.(string); ok {
			config.PublicAddress = c
		}
	}
	if opt, ok := opts["kubeconfig"]; ok {
		if c, ok := opt.(string); ok {
			config.KubeConfig = c
		}
	}
	if opt, ok := opts["namespace"]; ok {
		if c, ok := opt.(string); ok {
			config.Namespace = c
		}
	}
	if opt, ok := opts["in_cluster"]; ok {
		if c, ok := opt.(bool); ok {
			config.InCluster = c
		}
	}
	return config
}

// Constructor for kubernetesManager, takes options described in type config
func NewManager(c Config) (model.PlatformManager, error) {
	if c.PublicBindType == "" {
		c.PublicBindType = defaultPublicBindType
	}
	if c.PublicAddress == "" {
		c.PublicAddress = defaultPublicAddress
	}
	if c.KubeConfig == "" {
		c.KubeConfig = defaultKubeConfig
	}
	if c.Namespace == "" {
		c.Namespace = defaultNamespace
	}
	// Validation of Config
	switch c.PublicBindType {
	case "NodePort":
		// select handler funcs for nodeport
	default:
		return nil, fmt.Errorf("public_address_bind_type: unsupported value %s", c.PublicBindType)
	}
	var (
		kubeConf *rest.Config
		err error
	)
	if c.InCluster {
		kubeConf, err = rest.InClusterConfig()
		if err != nil {
			return nil, err
		}
	} else {
		kubeConf, err = clientcmd.BuildConfigFromFlags("", c.KubeConfig)
		if err != nil {
			return nil, err
		}
	}
	// create client
	client, err := kubernetes.NewForConfig(kubeConf)
	if err != nil {
		return nil, err
	}

	manager := &kubernetesManager{c: client, config: c}
	return manager, nil
}

func (m *kubernetesManager) Close() error {
	return nil
}
