// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package kubernetes

import (
	"fmt"
	"github.com/pkg/errors"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	appsv1 "k8s.io/api/apps/v1"
	batchv1 "k8s.io/api/batch/v1"
	apiv1 "k8s.io/api/core/v1"
	v1 "k8s.io/api/core/v1"
	kubeErrors "k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/util/intstr"
	"strconv"
	"time"
)

const ExternalSvcSuffix = "-external"

func (m *kubernetesManager) StartModule(module *model.Module, definition *model.ModuleDefinition, env map[string]string) (*model.Module, error) {
	m.sync.Lock()
	defer m.sync.Unlock()
	wrapErr := "cannot start module"
	var err error
	if module == nil {
		return nil, errors.Wrap(errors.New("received nil value"), wrapErr)
	}
	envVars := makeEnv(env)

	moduleType := definition.Type
	switch moduleType {
	case model.ModuleTypeService:
		if _, err := m.getDeployment(module.ID); err == nil {
			return module, errors.Wrap(fmt.Errorf("%s already running", module.ID), wrapErr)
		}

		portsExternal, portsInternal := makePorts(module.Ports)

		deployment := appsv1.Deployment{
			ObjectMeta: metav1.ObjectMeta{
				Name: module.ID,
			},
			Spec: appsv1.DeploymentSpec{
				Selector: &metav1.LabelSelector{
					MatchLabels: map[string]string{"name": module.ID},
				},
				Template: apiv1.PodTemplateSpec{
					Spec: apiv1.PodSpec{
						Containers: []apiv1.Container{
							{
								Name:  module.ID,
								Image: definition.Image + ":" + definition.Tag,
								Env:   envVars,
								LivenessProbe: makeStartupProbe(definition.HealthPort),
								Resources: v1.ResourceRequirements{
									Requests: makeResourceConstraints(definition.Resources),
								},
							},
						},
					},
					ObjectMeta: metav1.ObjectMeta{
						Name:   module.ID,
						Labels: map[string]string{"name": module.ID},
					},
				},
			},
		}
		err := m.createDeployment(&deployment)
		if err != nil {
			return module, errors.Wrap(err, wrapErr)
		}
		// put services to expose internal ports
		if len(portsInternal) > 0 {
			svcInternal := v1.Service{
				ObjectMeta: metav1.ObjectMeta{
					Name: module.ID,
				},
				Spec: v1.ServiceSpec{
					Type:     v1.ServiceType("ClusterIP"),
					Ports:    portsInternal,
					Selector: map[string]string{"name": module.ID},
				},
			}
			_, err = m.createService(&svcInternal)
			if err != nil {
				return module, errors.Wrap(err, wrapErr)
			}
		}
		// put services to expose public ports
		if len(portsExternal) > 0 {
			svcExternal := v1.Service{
				ObjectMeta: metav1.ObjectMeta{
					Name: module.ID + ExternalSvcSuffix,
				},
				Spec: v1.ServiceSpec{
					Type:     v1.ServiceType("NodePort"),
					Ports:    portsExternal,
					Selector: map[string]string{"name": module.ID},
				},
			}
			p, err := m.createService(&svcExternal)
			if err != nil {
				return module, errors.Wrap(err, wrapErr)
			}
			nodePorts := make(map[string]int)
			for _, port := range p.Spec.Ports {
				nodePorts[port.Name] = int(port.NodePort)
			}
			for i, port := range module.Ports {
				module.Ports[i].PublicEndpoint = m.config.PublicAddress + ":" + strconv.Itoa(nodePorts[port.Name])
			}
		}

	case model.ModuleTypeJob:
		if _, err := m.getJob(module.ID); err == nil {
			return module, errors.Wrap(fmt.Errorf("%s already running", module.ID), wrapErr)
		}
		job := batchv1.Job{
			ObjectMeta: metav1.ObjectMeta{
				Name: module.ID,
			},
			Spec: batchv1.JobSpec{
				Template: apiv1.PodTemplateSpec{
					Spec: apiv1.PodSpec{
						RestartPolicy: "OnFailure",
						Containers: []apiv1.Container{
							{
								Name:  module.ID,
								Image: definition.Image + ":" + definition.Tag,
								Env:   envVars,
								LivenessProbe: makeStartupProbe(definition.HealthPort),
								Resources: v1.ResourceRequirements{
									Requests: makeResourceConstraints(definition.Resources),
								},
							},
						},
					},
					ObjectMeta: metav1.ObjectMeta{
						Name:   module.ID,
						Labels: map[string]string{"name": module.ID},
					},
				},
			},
		}
		err = m.createJob(&job)
		if err != nil {
			return module, errors.Wrap(err, wrapErr)
		}
	default:
		return module, errors.Wrap(errors.New("unsupported module type"), wrapErr)
	}
	return module, nil
}

func (m *kubernetesManager) StopModule(moduleID string) error {
	m.sync.Lock()
	defer m.sync.Unlock()
	wrapErr := "cannot stop module"
	var err error
	var typ model.ModuleType

	if moduleID == "" {
		err = m.deleteAllJobs()
		if err != nil {
			return err
		}
		err = m.deleteAllDeployments()
		return err
	}

	if _, ok := m.getJob(moduleID); ok == nil {
		typ = model.ModuleTypeJob
	} else if _, ok := m.getDeployment(moduleID); ok == nil {
		typ = model.ModuleTypeService
	} else {
		// no module found
		return nil
	}
	switch typ {
	case model.ModuleTypeService:
		err = m.deleteDeployment(moduleID)
		if err != nil && !kubeErrors.IsNotFound(err) {
			return errors.Wrap(err, wrapErr)
		}
		err = m.deleteService(moduleID)
		if err != nil && !kubeErrors.IsNotFound(err) {
			return errors.Wrap(err, wrapErr)
		}
		err = m.deleteService(moduleID + ExternalSvcSuffix)
		if err != nil && !kubeErrors.IsNotFound(err) {
			return errors.Wrap(err, wrapErr)
		}
	case model.ModuleTypeJob:
		err = m.deleteJob(moduleID)
		if err != nil && !kubeErrors.IsNotFound(err) {
			return errors.Wrap(err, wrapErr)
		}
		// bugfix removed jobs still leave pods
		_ = m.deletePod(moduleID)
	}
	return nil
}


func (m *kubernetesManager) StatusModule(moduleID string) (model.ModuleStatus, error) {
	m.sync.Lock()
	defer m.sync.Unlock()
	wrapMsg := "cannot get module status"
	pod, err := m.c.CoreV1().Pods(m.config.Namespace).List(metav1.ListOptions{LabelSelector: moduleID})
	for count := 0; (err != nil || pod == nil || len(pod.Items) < 1) && count < 4; count += 1 {
		time.Sleep(time.Millisecond * 200)
		pod, err = m.c.CoreV1().Pods(m.config.Namespace).List(metav1.ListOptions{LabelSelector: "name=" + moduleID})
	}
	if err != nil || pod == nil || len(pod.Items) < 1 {
		return model.ModuleStatusUnknown, errors.Wrap(err, wrapMsg)
	}
	status := string(pod.Items[0].Status.Phase)
	// todo test
	switch status {
	case "Failed":
		return model.ModuleStatusFailed, nil
	case "Succeeded":
		return model.ModuleStatusCompleted, nil
	case "Pending":
		if len(pod.Items) != 0 && len(pod.Items[0].Status.ContainerStatuses) != 0 && pod.Items[0].Status.ContainerStatuses[0].State.Waiting != nil &&
			(pod.Items[0].Status.ContainerStatuses[0].State.Waiting.Reason == "ImagePullBackOff" || pod.Items[0].Status.ContainerStatuses[0].State.Waiting.Reason == "CrashLoopBackOff") {
			return model.ModuleStatusFailed, nil
		}
		return model.ModuleStatusCreating, nil
	case "Running":
		if len(pod.Items) != 0 && len(pod.Items[0].Status.ContainerStatuses) != 0 && pod.Items[0].Status.ContainerStatuses[0].State.Waiting != nil &&
			(pod.Items[0].Status.ContainerStatuses[0].State.Waiting.Reason == "ImagePullBackOff" || pod.Items[0].Status.ContainerStatuses[0].State.Waiting.Reason == "CrashLoopBackOff") {
			return model.ModuleStatusFailed, nil
		}
		return model.ModuleStatusRunning, nil
	}
	return model.ModuleStatusUnknown, nil
}

func makePorts(modulePorts []model.ModulePort) ([]v1.ServicePort, []v1.ServicePort) {
	portsInternal := make([]v1.ServicePort, 0)
	portsExternal := make([]v1.ServicePort, 0)
	for _, p := range modulePorts {
		if p.IsPublic {
			portsExternal = append(portsExternal, v1.ServicePort{Name: p.Name, Port: int32(p.Number)})
		} else {
			portsInternal = append(portsInternal, v1.ServicePort{Name: p.Name, Port: int32(p.Number)})
		}
	}
	return portsExternal, portsInternal
}

func makeEnv(vars map[string]string) []v1.EnvVar {
	env := make([]v1.EnvVar, 0)
	for name, value := range vars {
		env = append(env, v1.EnvVar{
			Name: name,
			Value: value,
		})
	}
	return env
}

func makeStartupProbe(port int) *v1.Probe {
	if port == 0 { return nil }
	return &v1.Probe{
		Handler: v1.Handler{
			HTTPGet:   &v1.HTTPGetAction{
				Path:        "/",
				Port:        intstr.FromInt(port),
			},
		},
		PeriodSeconds:       5,
		FailureThreshold:    100,
	}
}

func makeResourceConstraints(resources model.ModuleDefinitionResources) v1.ResourceList {
	resList := make(v1.ResourceList)
	if resources.Cpus != 0 {
		resList[v1.ResourceCPU] = resource.MustParse(strconv.Itoa(resources.Cpus))
	}
	if resources.Ram != 0 {
		resList[v1.ResourceMemory] = resource.MustParse(strconv.Itoa(resources.Ram) + "Mi")
	}
	if len(resList) == 0 {
		return nil
	}
	return resList
}