// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package mock

import (
	"errors"
	"fmt"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
	"sync"
	"time"
)

const timeServiceCreating time.Duration = time.Second * 4
const timeJobCreating time.Duration = time.Second * 4
const timeJobRunning time.Duration = time.Second * 14

type mockManager struct {
	sync sync.Mutex
	services map[string]time.Time
	jobs     map[string]time.Time
}

func NewMockManager() model.PlatformManager {
	return &mockManager{
		services: make(map[string]time.Time),
		jobs:     make(map[string]time.Time),
	}
}

func (m *mockManager) StartModule(module *model.Module, definition *model.ModuleDefinition, env map[string]string) (*model.Module, error) {
	m.sync.Lock()
	defer m.sync.Unlock()
	if _, ok := m.services[module.ID]; ok {
		return nil, errors.New("")
	}
	if _, ok := m.jobs[module.ID]; ok {
		return nil, errors.New("")
	}
	if definition.Type == model.ModuleTypeJob {
			m.jobs[module.ID] = time.Now()
	}
	if definition.Type == model.ModuleTypeService {
			m.services[module.ID] = time.Now()
	}
	for _, p := range module.Ports {
		if p.IsPublic {
			p.PublicEndpoint = "endpoint:9999"
		}
	}
	return module, nil
}

func (m *mockManager) StopModule(moduleID string) error {
	m.sync.Lock()
	defer m.sync.Unlock()
	if _, ok := m.services[moduleID]; ok {
		delete(m.services, moduleID)
		return nil
	}
	if _, ok := m.jobs[moduleID]; ok {
		delete(m.jobs, moduleID)
		return nil
	}
	return fmt.Errorf("no running module with ID %s", moduleID)
}

func (m *mockManager) StatusModule(moduleID string) (model.ModuleStatus, error) {
	if t, ok := m.services[moduleID]; ok {
		passed := time.Since(t)
		if passed < timeServiceCreating {
			return model.ModuleStatusCreating, nil
		} else  {
			return model.ModuleStatusRunning, nil
		}
	}
	if t, ok := m.jobs[moduleID]; ok {
		passed := time.Since(t)
		if passed < timeJobCreating {
			return model.ModuleStatusCreating, nil
		} else if passed > timeJobRunning  {
			return model.ModuleStatusCompleted, nil
		} else {
			return model.ModuleStatusRunning, nil
		}
	}
	return model.ModuleStatusUnknown, fmt.Errorf("no running module with ID %s", moduleID)
}

func (m *mockManager) Close() error {
	return nil
}