// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package repository

import (
	"github.com/jinzhu/gorm"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
)

// definitionRepository provides methods for storage handling of definition structures.
type definitionRepository struct {
	store *gorm.DB
}

// Takes existing Connection and returns new definition fetching repository
func NewDefinitionFetcher(c Connection) model.DefinitionFetcher {
	return newDefinitionRepository(c)
}

// Takes existing Connection and returns new definition storing repository
func NewDefinitionStorer(c Connection) model.DefinitionStorer {
	return newDefinitionRepository(c)
}

func newDefinitionRepository(c Connection) *definitionRepository {
	conn, _ := c.(*gormConnection)
	conn.db.AutoMigrate(
		&model.ModuleDefinitionPort{},
		&model.ModuleDefinitionParameter{},
		&model.ModuleDefinitionResources{},
		&model.ModuleDefinition{},
		&model.NodeDefinition{},
		&model.NodeDefinitionParameter{},
		&model.EdgeDefinition{},
		&model.OperationDefinition{},
	)
	return &definitionRepository{store: conn.db}
}

func (r *definitionRepository) StoreOperation(objs ...*model.OperationDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Create(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) StoreNode(objs ...*model.NodeDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Create(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) StoreEdge(objs ...*model.EdgeDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Create(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) StoreModule(objs ...*model.ModuleDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Create(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) UpdateOperation(obj *model.OperationDefinition) error {
	return r.store.Update(obj).Error
}

func (r *definitionRepository) UpdateNode(obj *model.NodeDefinition) error {
	return r.store.Update(obj).Error
}

func (r *definitionRepository) UpdateEdge(obj *model.EdgeDefinition) error {
	return r.store.Update(obj).Error
}

func (r *definitionRepository) UpdateModule(obj *model.ModuleDefinition) error {
	return r.store.Update(obj).Error
}

func (r *definitionRepository) DeleteOperation(objs ...*model.OperationDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) DeleteModule(objs ...*model.ModuleDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) DeleteNode(objs ...*model.NodeDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) DeleteEdge(objs ...*model.EdgeDefinition) error {
	return r.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (r *definitionRepository) QueryOperation(query *model.OperationDefinition, limit int) ([]*model.OperationDefinition, error) {
	m := make([]*model.OperationDefinition, 0)
	tx := r.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (r *definitionRepository) QueryNode(query *model.NodeDefinition, limit int) ([]*model.NodeDefinition, error) {
	m := make([]*model.NodeDefinition, 0)
	tx := r.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (r *definitionRepository) QueryEdge(query *model.EdgeDefinition, limit int) ([]*model.EdgeDefinition, error) {
	m := make([]*model.EdgeDefinition, 0)
	tx := r.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (r *definitionRepository) QueryModule(query *model.ModuleDefinition, limit int) ([]*model.ModuleDefinition, error) {
	m := make([]*model.ModuleDefinition, 0)
	tx := r.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}
