// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package repository

import (
	"github.com/jinzhu/gorm"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/model"
)

// metadataRepository provides methods for storage handling of metadata structures.
type metadataRepository struct {
	store *gorm.DB
}

// Takes existing Connection and returns new metadata repository
func NewMetadataRepository(c Connection) model.MetadataManager {
	conn, _ := c.(*gormConnection)
	conn.db.AutoMigrate(
		&model.Operation{},
		&model.OperationParameter{},
		&model.Node{},
		&model.Edge{},
		&model.Module{},
		&model.ModuleParameter{},
		&model.ModulePort{},
	)
	return &metadataRepository{conn.db}
}

func (s *metadataRepository) StoreOperation(obj *model.Operation) (*model.Operation, error) {
	err := s.store.Create(obj).Error
	return obj, err
}

func (s *metadataRepository) StoreNode(obj *model.Node) (*model.Node, error) {
	err := s.store.Create(obj).Error
	return obj, err
}

func (s *metadataRepository) StoreEdge(obj *model.Edge) (*model.Edge, error) {
	err := s.store.Create(obj).Error
	return obj, err
}

func (s *metadataRepository) StoreModule(obj *model.Module) (*model.Module, error) {
	err := s.store.Create(obj).Error
	return obj, err
}

func (s *metadataRepository) UpdateOperation(obj *model.Operation) error {
	return s.store.Model(&obj).Updates(obj).Error
}

func (s *metadataRepository) UpdateNode(obj *model.Node) error {
	return s.store.Model(&obj).Updates(obj).Error
}

func (s *metadataRepository) UpdateEdge(obj *model.Edge) error {
	return s.store.Model(&obj).Updates(obj).Error
}

func (s *metadataRepository) UpdateModule(obj *model.Module) error {
	return s.store.Model(&obj).Updates(obj).Error
}

func (s *metadataRepository) DeleteOperation(objs ...*model.Operation) error {
	return s.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (s *metadataRepository) DeleteNode(objs ...*model.Node) error {
	return s.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (s *metadataRepository) DeleteEdge(objs ...*model.Edge) error {
	return s.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (s *metadataRepository) DeleteModule(objs ...*model.Module) error {
	return s.store.Transaction(func(tx *gorm.DB) error {
		for _, obj := range objs {
			if err := tx.Delete(obj).Error; err != nil {
				// TODO err handle
				return err
			}
		}
		return nil
	})
}

func (s *metadataRepository) QueryOperation(query *model.Operation, limit int) ([]*model.Operation, error) {
	m := make([]*model.Operation, 0)
	tx := s.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (s *metadataRepository) QueryNode(query *model.Node, limit int) ([]*model.Node, error) {
	m := make([]*model.Node, 0)
	tx := s.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (s *metadataRepository) QueryEdge(query *model.Edge, limit int) ([]*model.Edge, error) {
	m := make([]*model.Edge, 0)
	tx := s.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}

func (s *metadataRepository) QueryModule(query *model.Module, limit int) ([]*model.Module, error) {
	m := make([]*model.Module, 0)
	tx := s.store.Set("gorm:auto_preload", true)
	if limit != 0 {
		tx = tx.Limit(limit)
	}
	if query != nil {
		tx = tx.Where(query)
	}
	err := tx.Find(&m).Error
	if err != nil {
		// TODO err handle
		return nil, err
	}
	return m, nil
}
