// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Package repository provides functions to create new Metadata or Definition repositories
// provided by one of supported database drivers.
// Currently supported database drivers are: "sqlite3", "postgres", "mssql", "mysql".
package repository

import (
	"database/sql"
	"github.com/jinzhu/gorm"
	_ "github.com/jinzhu/gorm/dialects/mssql"
	_ "github.com/jinzhu/gorm/dialects/mysql"
	_ "github.com/jinzhu/gorm/dialects/postgres"
	_ "github.com/jinzhu/gorm/dialects/sqlite"
)

// Connection represents active connection to sql database. Only functions from this package can make
// Connections used in initialising repositories to ensure correct frameworks are in use.
type Connection interface {
	// to ensure interface will no be satisfied by anything from outside
	private()
	// Returns raw sql connection
	Get() *sql.DB
}

// implementation of connection using only supported gorm framework
type gormConnection struct {
	db *gorm.DB
}

// to satisfy interface
func (c *gormConnection) private() {}

func (c *gormConnection) Get() *sql.DB {
	return c.db.DB()
}

// Represents configuration options for repository.
// Not all options are needed for particular driver e.g. sqlite3 does not need any options specified.
type Options struct {
	// Database driver to use.
	// Currently supported database drivers are: "sqlite3", "postgres", "mssql", "mysql".
	Driver string
	// Hostname or IP address of database location
	Host string
	// Network port of database location
	Port string
	// Database name
	Name string
	// Username
	User string
	// Password
	Password string
}

// Creates connection to sql database using gorm framework.
// Returns error if database type not supported or connection cannot be established
func NewConnection(c Options) (Connection, error) {
	// default values
	const (
		defaultMysqlPort = "3306"
		defaultPostgresPort = "5432"
		defaultMssqlPort = "1433"
		defaultHost = "127.0.0.1"
	)
	// use default host if none provided
	if c.Host == "" {
		c.Host = defaultHost
	}
	var options string
	switch c.Driver {
	case "mysql":
		if c.Port == "" {
			c.Port = defaultMysqlPort
		}
		options = c.User +":" + c.Password + "@(" + c.Host + ":" + c.Port + ")/"+ c.Name + "?charset=utf8&parseTime=True&loc=Local"
	case "postgres":
		if c.Port == "" {
			c.Port = defaultPostgresPort
		}
		options = "host=" + c.Host + " port=" + c.Port + " user=" + c.User + " dbname=" + c.Name + " password=" + c.Password
	case "mssql":
		if c.Port == "" {
			c.Port = defaultMssqlPort
		}
		options = "sqlserver://"+ c.User + ":" + c.Name + "@" + c.Host + ":" + c.Port + "?database=" + c.Name
	case "sqlite3":
		// TODO support write to file
		options = "file::memory:?cache=shared"
	}
	db, err := gorm.Open(c.Driver, options)
	return &gormConnection{db: db}, err
}

