// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package mEngine

// Package mEngine implements semi-parallel processing of requests consisting of 1 Task and a number of blocking Dependencies.
// Task can be added into group. Tasks in group are processed in sequence.
// However, each group is processed independently of each other.

import (
	"log"
	"sync"
	"time"
)

// Type EngineOption represents option for Engine
type EngineOption func(*mEngine)

// FlushPeriod sets time period of flushing already finished runners and blockers
func FlushPeriod(period time.Duration) EngineOption {
	return func(e *mEngine) {
		e.flushPeriod = period
	}
}

// NewEngine initializes new precessing engine
func NewEngine(opts ...EngineOption) Engine {
	const (
		defaultFlushPeriod = time.Second * 15
	)
	m := mEngine{
		flushPeriod: defaultFlushPeriod,
		requestQueue: make(chan *req, 30),
		finished: make(chan struct{}),
		recentBlockers: make(map[string][]*blocker),
		recentRunners: make(map[string]*runner),
		openForRequests: make(chan struct{}),
	}
	close(m.openForRequests)
	for _, opt := range opts {
		opt(&m)
	}
	go m.loop()
	return &m
}

// Type Runner represents a function that can be processed by Engine.
//Its length of run should be little as it cannot get canceled.
type Runner func(DoneFunc)

// Type Blocker represents a function that block Runner from start until Blocker finishes
// Blockers can run for a longer time as they can be canceled.
type Blocker func(StopFunc, DoneFunc)

// Type StopFunc is a function that returns channel, which gets closed when Engine want to cancel Blocker,
type StopFunc func() <-chan struct{}
// Type DoneFunc is a function that returns channel, which should get closed in Runner/Blocker when its finished
type DoneFunc func() chan<- struct{}

// type Engine represents an Engine for processing runners and their blockers.
type Engine interface {
	NewTask(group string, runner Runner, blockers ...Blocker)
	NewTaskWithCancel(group string, runner Runner, blockers ...Blocker)
	Close()
}

type blocker struct {
	runFn Blocker
	stopChannel chan struct{}
	doneChannel chan struct{}
}

func (b *blocker) cancel() {
	select {
		case b.stopChannel <- struct{}{}:
		default:
	}
}

func (b *blocker) finished() <-chan struct{} {
	return b.doneChannel
}

func (b *blocker) start() {
	stop := func() <-chan struct{} { return b.stopChannel }
	done := func() chan<- struct{} {
		return b.doneChannel
	}
	go b.runFn(stop, done)
}

// type Runner
type runner struct {
	runFn Runner
	doneChannel chan struct{}
	started bool
}

func (t *runner) finished() <-chan struct{} {
	return t.doneChannel
}

func (t *runner) start() {
	if t.started {
		log.Fatal()
	}
	t.started = true
	done := func() chan<- struct{} { return t.doneChannel }
	go t.runFn(done)
}

// Type Request
type req struct {
	group string
	cancelBlockers bool
	runner *runner
	blockers []*blocker
}

type mEngine struct {
	flushPeriod time.Duration

	recentRunners map[string]*runner
	recentBlockers map[string][]*blocker

	requestQueue chan *req

	openForRequests chan struct{}
	finished chan struct{}
}

func (e *mEngine) request(group string, r Runner, b ...Blocker) req {
	t := runner{
		runFn: r,
		doneChannel: make(chan struct{}),
	}
	blockers := make([]*blocker, 0)
	for _, fn := range b {
		blocker := blocker{
			runFn: fn,
			stopChannel: make(chan struct{}, 1),
			doneChannel: make(chan struct{}),
		}
		blockers = append(blockers, &blocker)
	}
	return req{
		group: group,
		runner: &t,
		blockers: blockers,
		cancelBlockers: false,
	}
}

// Close finishes processing tasks and tear down Engine
func (e *mEngine) Close() {
	e.openForRequests = nil
	// to mitigate inconsistencies
	<- time.After(time.Millisecond * 200)
	close(e.requestQueue)
	<- e.finished
	return
}

// NewTask sends runner and number of its blockers to processing queue
func (e *mEngine) NewTask(group string, runner Runner, blockers ...Blocker) {
	r := e.request(group, runner, blockers...)
	select {
	case <- e.openForRequests:
		e.requestQueue <- &r
	default:
	}
}

// NewTaskWithCancel sends runner and number of its blockers to processing queue.
// However, it also cancels out any remaining blockers in the group
func (e *mEngine) NewTaskWithCancel(group string, runner Runner, blockers ...Blocker) {
	r := e.request(group, runner, blockers...)
	r.cancelBlockers = true
	select {
	case <- e.openForRequests:
		e.requestQueue <- &r
	default:
	}
}

// flush all runners and blockers stored in Engine, that are already done running
func (e *mEngine) flush() {
	// flush runners
	for k, v := range e.recentRunners {
		select {
		case <- v.finished():
			delete(e.recentRunners, k)
		default:
			continue
		}
	}
	for k, blockers := range e.recentBlockers {
		// flush blockers
		temp := blockers[:0]
		for _, b := range blockers {
			select {
			case <- b.finished():
				continue
			default:
				temp = append(temp, b)
			}
		}
		if len(temp) == 0 {
			delete(e.recentBlockers, k)
		} else {
			e.recentBlockers[k] = temp
		}
	}
}

func (e *mEngine) runRequest(request *req) {
	if request.cancelBlockers {
		if blockers, ok := e.recentBlockers[request.group]; ok {
			for _, b := range blockers {
				b.cancel()
			}
		}
	}
	// run all blockers of request and put them to list of active blockers
	for _, b := range request.blockers {
		b.start()
	}
	if e.recentBlockers[request.group] == nil {
		e.recentBlockers[request.group] = make([]*blocker, 0)
	}
	e.recentBlockers[request.group] = append(e.recentBlockers[request.group], request.blockers...)
	// check if runner with same ID already in queue
	previous, ok := e.recentRunners[request.group]
	// attach itself as queued runner
	e.recentRunners[request.group] = request.runner
	go func() {
		// wait for queued runner to complete run
		if ok {
			<- previous.finished()
		}
		var wg sync.WaitGroup
		// wait for all blockers to complete
		for _, b := range request.blockers {
			b := b
			wg.Add(1)
			go func() {
				<- b.finished()
				wg.Done()
			}()
		}
		wg.Wait()
		// run runner
		request.runner.start()
	}()
}

func (e *mEngine) teardown() {
	// close all blockers
	for _, v := range e.recentBlockers {
		for _, b := range v {
			b.cancel()
		}
	}
	// wait for all runners to finish
	var wg sync.WaitGroup
	for _, r := range e.recentRunners {
		wg.Add(1)
		c := r.finished()
		go func(){
			<- c
			wg.Done()
		}()
	}
	wg.Wait()
}

func (e *mEngine) loop() {
	// do all requests until channel closed
	ticker := time.NewTicker(e.flushPeriod)
	for {
		select {
		case m, ok := <- e.requestQueue:
			// tear-down and finish
			if !ok {
				e.teardown()
				ticker.Stop()
				close(e.finished)
				return
			}
			// start all blockers and runner in separate goroutine
			e.runRequest(m)
			// process request
		case <- ticker.C:
			// flush not needed
			e.flush()
		}
	}
}

