// Copyright (C) 2020 Masaryk University
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package cmd

import (
	"context"
	"fmt"
	"github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/app"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/config"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/logger"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/platform"
	"gitlab.ics.muni.cz/ANALYZA/orchestrator/internal/repository"
	"net/http"
	"os"
	"os/signal"
	"path"
	"time"
)

func Execute() {
	rootCmd.PersistentFlags().BoolP("debug", "v", false, "enables debug mode")
	rootCmd.PersistentFlags().StringP("config", "c", "", "specifies path to config file")
	rootCmd.PersistentFlags().BoolP("log", "l", false, "enables logging to console")
	rootCmd.AddCommand(versionCmd)
	if err := rootCmd.Execute(); err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
}

var rootCmd = &cobra.Command{
	Use:   "orchestrator",
	Short: "orchestrator manages kubernetes resources, maintains database and provide rest API for the gui part of Analyza project.",
	RunE: func(cmd *cobra.Command, args []string) error {
		var err error
		// get config struct
		configPath, _ :=  cmd.Flags().GetString("config")
		config, err := config.GetConfig(configPath)
		if err != nil {
			return err
		}
		// add values from flags
		debug, _ := cmd.Flags().GetBool("debug")
		if debug {
			config.Log.Debug = debug
		}
		stdLog, _ := cmd.Flags().GetBool("log")

		// setup logging
		var logFile *os.File
		var logFormat logrus.Formatter
		// prepare log file and format
		if stdLog {
			logFile = os.Stdout
			logFormat = &logrus.TextFormatter{
				FullTimestamp:   false,
				QuoteEmptyFields: true,
				TimestampFormat: time.RFC3339Nano,
			}
		} else {
			if err := os.MkdirAll(path.Dir(config.Log.Logfile), os.ModeDir | 0700); err != nil {
				return err
			}
			logFile, err = os.Create(config.Log.Logfile)
			if err != nil {
				return err
			}
			defer logFile.Close()
			logFormat = &logrus.JSONFormatter{}
		}
		// set log level
		var logLevel logrus.Level
		if config.Log.Debug {
			logLevel = logrus.DebugLevel
		} else {
			logLevel = logrus.InfoLevel
		}
		// initialize logger
		logger := logger.New(logLevel, logFormat, logFile)

		// initialise platform manager
		platformManager, err := platform.NewPlatformManager(config.Platform.PlatformType, config.Platform.Options)
		if err != nil {
			logger.Log("error", err)
			return err
		}
		// ensure platform manager will close connection at the end
		defer func() {
			err = platformManager.Close()
			if err != nil {
				logger.Log("error", fmt.Errorf("error in closing connection to platform: %s", err))
			}
		}()

		// initialise definition manager
		connectionDef, err := repository.NewConnection(repository.Options{
			Driver:     config.Definitions.DBType,
			Host:       config.Definitions.Options.Host,
			Port:       config.Definitions.Options.Port,
			Name:       config.Definitions.Options.Database,
			User:       config.Definitions.Options.User,
			Password:   config.Definitions.Options.Password,
		})
		if err != nil {
			logger.Log("error", err)
			return err
		}
		definitionManager := repository.NewDefinitionFetcher(connectionDef)
		// initialise metadata manager
		connectionMeta , err := repository.NewConnection(repository.Options{
			Driver:     config.Metadata.DBType,
			Host:       config.Metadata.Options.Host,
			Port:       config.Metadata.Options.Port,
			Name:       config.Metadata.Options.Database,
			User:       config.Metadata.Options.User,
			Password:   config.Metadata.Options.Password,
		})
		if err != nil {
			logger.Log("error", err)
			return err
		}
		metadataManager := repository.NewMetadataRepository(connectionMeta)
		// initialize orchestrator app
		orchestrator, err := app.NewOrchestrator(app.Options{
			WareHouse:  config.Warehouse.Host + ":" + config.Warehouse.Port,
			DockerRegistry: config.Docker.Registry,
			Definition: definitionManager,
			Metadata:   metadataManager,
			Platform:   platformManager,
			Logger:     logger,
		})
		if err != nil {
			logger.Log("error", err)
			return err
		}
		// ensure orchestrator will finish work at the end
		defer orchestrator.Close()

		// get routes
		router := orchestrator.GetRoutes()
		server := &http.Server{
			Addr: fmt.Sprintf("%s:%s", config.Api.Address, config.Api.Port),
			Handler: router,
		}

		// run http server in goroutine
		finishServer := make(chan struct{})
		go func() {
			defer close(finishServer) // let main know we are done cleaning up
			// always returns error. ErrServerClosed on graceful close
			if err := server.ListenAndServe(); err != http.ErrServerClosed {
				// unexpected error
				logger.Log("error", err)
			}
		}()

		// initialise waiter for os interrupt
		finishCatcher := make(chan struct{})
		go func() {
			defer close(finishCatcher)
			ch := make(chan os.Signal, 1)
			signal.Notify(ch, os.Interrupt)
			select {
			// when interrupt caught, try to gracefully shutdown server
			case <- ch:
				logger.Log("info", "Orchestrator API closing...")
				ctx, _ := context.WithTimeout(context.Background(), time.Second * 5)
				if err := server.Shutdown(ctx); err != nil {
					logger.Log("error", err) // failure/timeout shutting down the server gracefully
				}
			// if server already finished, just return
			case <- finishServer:
			}
		}()
		logger.Log("info", fmt.Sprintf("Orchestrator API listening on %s:%s", config.Api.Address, config.Api.Port))
		// wait for server to finish
		<- finishServer
		<- finishCatcher
		return nil
	},
}
